"""
This collection of functions is used for general work on wavelength reduced échelle spectra.
"""
import numpy as np
from astropy.io import fits
import matplotlib.pyplot as plt
import scipy.constants as scc
from scipy.interpolate import interp1d
from numpy.polynomial import Polynomial
from matplotlib.collections import PathCollection
from matplotlib.legend_handler import HandlerPathCollection, HandlerLine2D
import matplotlib


def load_template_spectrum(template_spectrum_path: str):
    with fits.open(template_spectrum_path) as hdul:
        hdr = hdul[0].header
        flux = hdul[0].data
        wl0, delta_wl, naxis = hdr['CRVAL1'], hdr['CDELT1'], hdr['NAXIS1']
    wavelength = np.linspace(wl0, wl0+delta_wl*naxis, naxis)
    return wavelength, flux


def load_program_spectrum(program_spectrum_path: str):
    with fits.open(program_spectrum_path) as hdul:
        hdr = hdul[0].header
        flux = hdul[0].data
        wl0, delta_wl = hdr['CRVAL1'], hdr['CDELT1']
        RA, DEC = hdr['OBJRA'], hdr['OBJDEC']
        date = hdr['DATE-AVG']
    wavelength = np.linspace(wl0, wl0 + delta_wl * flux.size, flux.size)
    return wavelength, flux, date, RA, DEC


def _resample_multiple_spectra(wavelength_template: np.ndarray, wavelength_list: list, flux_list: list):

    flux_resampled_collection = np.empty(shape=(wavelength_template.size, len(flux_list)))
    for i in range(0, len(flux_list)):
        flux_resampled_collection[:, i] = interp1d(wavelength_list[i], flux_list[i], kind='linear')(wavelength_template)
        # flux_resampled_collection[:, i] = np.interp(wavelength_template, wavelength_list[i], flux_list[i])

    return flux_resampled_collection


def _create_wavelength_template(
        wavelengths: list or np.ndarray, delta_v:float,  wavelength_a:float = None, wavelength_b:float = None,
        even_length=True
):
    if isinstance(wavelengths, list):
        if wavelength_a is None:
            wavelength_a = np.max([np.min(x) for x in wavelengths])
        if wavelength_b is None:
            wavelength_b = np.min([np.max(x) for x in wavelengths])
    else:
        if wavelength_a is None:
            wavelength_a = np.min(wavelengths)
        if wavelength_b is None:
            wavelength_b = np.max(wavelengths)

    speed_of_light = scc.c / 1000  # in km/s
    step_amnt = int(np.log10(wavelength_b / wavelength_a) / np.log10(1.0 + delta_v / speed_of_light))
    wavelength_template = wavelength_a * (1.0 + delta_v / speed_of_light) ** (np.linspace(1, step_amnt, step_amnt))

    if even_length is True and np.mod(wavelength_template.size, 2) != 0.0:
        wavelength_template = wavelength_template[:-1]

    return wavelength_template


def resample_to_equal_velocity_steps(
        wavelength: np.ndarray or list, delta_v: float, flux=None, wavelength_template: np.ndarray = None,
        wavelength_a=None, wavelength_b=None, resampled_len_even=True
):
    """
    Multi-use function that can be used to either: Create a new wavelength grid equi-distant in vel
    (without interpolation), create grid and resample flux to it, create mutual grid and resample flux for multiple
    spectra simultanously. It can also resample one (or more) spectra to a provided wavelength grid.
    See interpolate_to_equal_velocity_steps() for simpler illustration of what this function does.
    :param wavelength:              either a list of np.ndarray's, or a np.ndarray
    :param delta_v:                 desired spectrum resolution in velocity space in km/s
    :param flux:                    either a list of np.ndarray's, or a np.ndarray. Set to None if only
                                    wavelength_template should be returned.
    :param wavelength_template:    a np.ndarray with a provided resampled grid. Set to None if one should be calculated
    :param wavelength_a:            float, start of desired grid. If None, calculates from provided spectra
    :param wavelength_b:            float, end of desired grid. If None, calculates from provided spectra.
    :param resampled_len_even:      bool switch, sets if resampled grid should be kept on an even length
    :return:    either wavelength_template,
                      (wavelength_template, flux_resampled_collection),
                or    (wavelength_template, flux_resampled)
    """
    if wavelength_template is None:
        wavelength_template = _create_wavelength_template(
           wavelength, delta_v, wavelength_a, wavelength_b, resampled_len_even
        )
    elif resampled_len_even and np.mod(wavelength_template.size, 2) != 0.0:
        wavelength_template = wavelength_template[:-1]

    if flux is not None:
        if isinstance(flux, list) and isinstance(wavelength, list):
            flux_resampled_collection = _resample_multiple_spectra(wavelength_template, wavelength, flux)
            return wavelength_template, flux_resampled_collection
        elif isinstance(flux, np.ndarray) and isinstance(wavelength, np.ndarray):
            flux_resampled = interp1d(wavelength, flux, kind='linear')(wavelength_template)
            # flux_resampled = np.interp(wavelength_template, wavelength, flux)
            return wavelength_template, flux_resampled
        else:
            raise ValueError("flux and wavelength is neither a list (of arrays), an array, or None "
                             "(they must be the same type).")
    else:
        return wavelength_template


def resample_multiple_spectra(
        delta_v: float, *spectra, wavelength_template: np.ndarray = None, wavelength_a=None, wavelength_b=None,
        resampled_len_even=True
):
    """
    Resamples multiple spectra. Expects *spectra (traditionally *args in python) to be a collection of tuples with
    (wavelength, flux) for each. It is allowed to include multiple spectra in one of the tuples if done with
    ([wl_1, wl_2, ...], [flux_1, flux_2, ...]) since this function calls resample_to_equal_velocity_steps(). As such
    this function is purely a wrapper for convenience.
    :return: wavelength, (flux_1, flux_2, flux_3, ...)
    """
    flux_list = []
    for (wl, fl) in spectra:
        wavelength_template, flux_ = resample_to_equal_velocity_steps(
            wl, delta_v, fl, wavelength_template, wavelength_a, wavelength_b, resampled_len_even
        )
        flux_list.append(flux_)
    return wavelength_template, tuple(flux_list)


def interpolate_to_equal_velocity_steps(wavelength_collector_list: list, flux_collector_list: list, delta_v: float):
    """
    Resamples a set of spectra to the same wavelength grid equi-spaced in velocity map.
    :param wavelength_collector_list:   list of arrays, one array for each spectrum
    :param flux_collector_list:         list of arrays, one array for each spectrum
    :param delta_v:                     interpolation resolution for spectrum in km/s
    :return: wavelength, flux_collection_array
    """
    speed_of_light = scc.c / 1000       # in km/s

    # # Create unified wavelength grid equispaced in velocity # #
    wavelength_a = np.max([x[0] for x in wavelength_collector_list])
    wavelength_b = np.min([x[-1] for x in wavelength_collector_list])
    step_amnt = np.log10(wavelength_b / wavelength_a) / np.log10(1.0 + delta_v / speed_of_light) + 1
    wavelength = wavelength_a * (1.0 + delta_v / speed_of_light) ** (np.linspace(1, step_amnt, step_amnt))

    # # Interpolate to unified wavelength grid # #
    flux_collection_array = np.empty(shape=(wavelength.size, len(flux_collector_list)))
    for i in range(0, len(flux_collector_list)):
        flux_interpolator = interp1d(wavelength_collector_list[i], flux_collector_list[i], kind='linear')
        flux_collection_array[:, i] = flux_interpolator(wavelength)

    return wavelength, flux_collection_array


def _create_buffer_mask(wavelength: np.ndarray, wavelength_limits: tuple, buffer_size: float):
    buffer_lower = (wavelength > wavelength_limits[0] - buffer_size) & (wavelength < wavelength_limits[0])
    buffer_upper = (wavelength < wavelength_limits[1] - buffer_size) & (wavelength > wavelength_limits[1])
    buffer_mask = buffer_lower | buffer_upper
    return buffer_mask


def limit_wavelength_interval(
        wavelength_limits: tuple, wavelength: np.ndarray, flux: np.ndarray, buffer_size=None, buffer_mask=None,
        even_length=False
):
    selection_mask = (wavelength > wavelength_limits[0]) & (wavelength < wavelength_limits[1])
    selection_mask_buffered = np.zeros(wavelength.shape, dtype=bool)
    if buffer_size is not None and buffer_mask is None:
        buffer_mask = _create_buffer_mask(wavelength, wavelength_limits, buffer_size)
    if buffer_mask is not None:
        selection_mask_buffered = selection_mask | buffer_mask

    wavelength_new = wavelength[selection_mask]
    if flux.ndim == 1:
        flux_new = flux[selection_mask]
    elif flux.ndim == 2:
        flux_new = flux[selection_mask, :]
    else:
        raise ValueError('flux has wrong number of dimensions. Should be either 1 or 2.')

    if buffer_mask is not None:
        wavelength_buffered = wavelength[selection_mask_buffered]
        if flux.ndim == 1:
            flux_buffered = flux[selection_mask_buffered]
        elif flux.ndim == 2:
            flux_buffered = flux[selection_mask_buffered, :]
        else:
            raise ValueError('flux has wrong number of dimensions. Should be either 1 or 2.')

    if even_length is True:
        wavelength_new, flux_new = make_spectrum_even(wavelength_new, flux_new)

        if np.mod(wavelength.size, 2) != 0.0:
            raise ValueError('Input wavelength must be even length for even buffer_mask.')
        elif buffer_mask is not None:
            wavelength_buffered, flux_buffered = make_spectrum_even(wavelength_buffered, flux_buffered)
            intermediate_array = wavelength_buffered[np.in1d(wavelength_buffered, wavelength_new, invert=True)]
            buffer_mask = np.in1d(wavelength, intermediate_array)

    if buffer_mask is not None:
        buffer_mask_new = np.in1d(wavelength_buffered, wavelength_new, invert=True)
        return (wavelength_new, flux_new), (wavelength_buffered, flux_buffered, buffer_mask_new, buffer_mask)
    else:
        return wavelength_new, flux_new


def limit_wavelength_interval_multiple_spectra(
        wavelength_limits: tuple, wavelength: np.ndarray, *args, buffer_size=None, even_length=True
):
    """
    Expects all *args to be fluxes, either a collection of spectral fluxes (np.ndarray with ndims 2) or a single
    spectrum each (np.ndarray with ndims 1). They must all have been resampled to the same wavelength grid beforehand.
    :return: lists of wavelength limited spectra, including their wavelength values. Will also return buffered versions
              (with padding in the ends) if buffer_size is given, as well as a buffer mask that can be used to access
              the buffer or the unbuffered data from it.
    """
    buffer_mask_internal = None
    flux_buffered_list = []
    flux_unbuffered_list = []
    for arg in args:
        temp_0, temp_1 = limit_wavelength_interval(
            wavelength_limits, wavelength, arg, buffer_size, buffer_mask_internal, even_length
        )
        if buffer_size is None:
            wavelength_unbuffered = temp_0
            flux_unbuffered_list.append(temp_1)
        else:
            wavelength_unbuffered = temp_0[0]
            flux_unbuffered_list.append(temp_0[1])
            wavelength_buffered = temp_1[0]
            flux_buffered_list.append(temp_1[1])
            buffer_mask_return = temp_1[2]
            buffer_mask_internal = temp_1[3]

    if buffer_size is None:
        return wavelength_unbuffered, flux_unbuffered_list
    else:
        return wavelength_unbuffered, flux_unbuffered_list, wavelength_buffered, flux_buffered_list, buffer_mask_return


def make_spectrum_even(wavelength:np.ndarray, flux:np.ndarray or list):
    if np.mod(wavelength.size, 2) != 0.0:
        wavelength = wavelength[:-1]
        if isinstance(flux, np.ndarray):
            if flux.ndim == 1:
                flux = flux[:-1]
            elif flux.ndim == 2:
                flux = flux[:-1, :]
            else:
                raise ValueError('flux dimension must be 1 or 2.')
        elif isinstance(flux, list):
            for i in range(0, len(flux)):
                if flux[i].ndim == 1:
                    flux[i] = flux[i][:-1]
                elif flux[i].ndim == 2:
                    flux[i] = flux[i][:-1, :]
                else:
                    raise ValueError('flux[i] dimension must be 1 or 2.')
    return wavelength, flux


def moving_median_filter(flux: np.ndarray, window=51):
    """
    Filters the data using a moving median filter. Useful to capture some continuum trends.
    :param flux:        np.ndarray size (n, ) of y-values (fluxes) for each wavelength
    :param window:      integer, size of moving median filter
    :return:            filtered_flux, np.ndarray size (n, )
    """
    from scipy.ndimage import median_filter
    return median_filter(flux, size=window, mode='reflect')


def reduce_emission_lines(wavelength: np.ndarray, flux: np.ndarray, mf_window=401, std_factor=1.5, plot=False,
                          limit=None):
    """
    Reduces the effect of emission lines in a spectrum by median-filtering and cutting away all data points a certain
    point ABOVE the median variance of the data (flux - median flux).
    :param wavelength:              np.ndarray size (n, ) of wavelengths.
    :param flux:                    np.ndarray size (n, ) of fluxes for the spectrum
    :param mf_window:               int, window size of the median filter. Must be uneven.
    :param std_factor:              float, cutoff factor multiplier to the standard deviation of the variance.
    :param plot:                    bool, indicates whether illustrative plots should be shown
    :param limit:                   float, max distance from variance allowed. Default is None
    :return wavelength_reduced:     np.ndarray of wavelengths, masked to the reduced fluxes.
            flux_emission_reduced:  np.ndarray of fluxes for the spectrum, with reduced emission lines.
    """
    flux_filtered = moving_median_filter(flux, mf_window)
    variance = flux - flux_filtered
    mask = variance < std_factor * np.std(variance)
    wavelength_reduced = wavelength[mask]
    flux_emission_reduced = flux[mask]
    if limit is not None:
        mask = variance[mask] < limit
        wavelength_reduced = wavelength_reduced[mask]
        flux_emission_reduced = flux_emission_reduced[mask]

    if plot:
        plt.figure()
        plt.plot(wavelength, flux); plt.plot(wavelength, flux_filtered)
        plt.xlabel('Wavelength'); plt.ylabel('Flux'); plt.legend(['Unfiltered', 'Median Filtered'])
        plt.show(block=False)

        plt.figure()
        plt.plot(wavelength, variance); plt.plot(wavelength_reduced, variance[mask])
        plt.plot([np.min(wavelength), np.max(wavelength)], [std_factor*np.std(variance), std_factor*np.std(variance)],
                 'k')
        plt.plot([np.min(wavelength), np.max(wavelength)],
                 [-std_factor * np.std(variance), -std_factor * np.std(variance)], 'k--')
        plt.xlabel('Wavelength'); plt.ylabel('Variance'); plt.legend(['Full set', 'Set with reduced emission lines'])
        plt.show(block=False)

        plt.figure()
        plt.plot(wavelength, flux, linewidth=1)
        plt.plot(wavelength_reduced, flux_emission_reduced, '--', linewidth=0.8)
        plt.xlabel('Wavelength'); plt.ylabel('Flux'); plt.legend(['Full set', 'Set with reduced emission lines'])
        plt.show()

    return wavelength_reduced, flux_emission_reduced


def updatescatter(handle, orig):
    """https://stackoverflow.com/questions/24706125/setting-a-fixed-size-for-points-in-legend"""
    handle.update_from(orig)
    handle.set_sizes([64])


def updateline(handle, orig):
    """https://stackoverflow.com/questions/24706125/setting-a-fixed-size-for-points-in-legend"""
    handle.update_from(orig)
    handle.set_markersize(12)


def simple_normalizer(wavelength, flux, reduce_em_lines=True, plot=False):
    matplotlib.rcParams.update({'font.size': 20})
    filtered_flux = moving_median_filter(flux, flux.size//25)

    selection_mask_1 = flux/filtered_flux > 0.85
    polynomial_1 = Polynomial.fit(wavelength[selection_mask_1], flux[selection_mask_1], deg=3)

    selection_mask_2 = (flux/polynomial_1(wavelength) > 0.95) & (flux/polynomial_1(wavelength) < 1.1)
    polynomial_2 = Polynomial.fit(wavelength[selection_mask_2], flux[selection_mask_2], deg=4)

    normalized_flux = flux / polynomial_2(wavelength)

    if reduce_em_lines:
        std = np.std(normalized_flux[selection_mask_1])
        mean = np.mean(normalized_flux[selection_mask_1])
        selection_mask_3 = normalized_flux < mean + 2.5*std

    if plot:
        plt.figure(figsize=(16, 9))
        plt.plot(wavelength, flux, '.', markersize=1)
        plt.plot(wavelength, filtered_flux)
        plt.plot(wavelength, polynomial_1(wavelength))
        plt.plot(wavelength, polynomial_2(wavelength))

        plt.xlabel('Wavelength [Å]', fontsize=22)
        plt.ylabel('Spectral Flux [e/s]', fontsize=22)
        if reduce_em_lines:
            plt.plot(wavelength[~selection_mask_3], flux[~selection_mask_3], '.', color='red', markersize=4)
            plt.legend(
                ['Flux', 'Median filtered flux', '3rd degree Polynomial', '4th degree polynomial', 'Removed data'],
                fontsize=20,
                handler_map={PathCollection: HandlerPathCollection(update_func=updatescatter),
                             plt.Line2D: HandlerLine2D(update_func=updateline)}
            )
        else:
            plt.legend(
                ['Flux', 'Median filtered flux', '3rd degree Polynomial', '4th degree polynomial'],
                fontsize=20,
                handler_map={PathCollection: HandlerPathCollection(update_func=updatescatter),
                             plt.Line2D: HandlerLine2D(update_func=updateline)}
            )
        plt.tight_layout()
        # plt.savefig('/home/sinkbaek/PycharmProjects/Seismic-dEBs/figures/report/RV/simple_normalizer/continuum_fit.png',
        #            dpi=400)

        plt.show(block=True)

    if reduce_em_lines:
        return wavelength[selection_mask_3], normalized_flux[selection_mask_3]
    else:
        return wavelength, normalized_flux


def save2col(column1: np.ndarray, column2: np.ndarray, filename: str):
    save_data = np.empty((column1.size, 2))
    save_data[:, 0] = column1
    save_data[:, 1] = column2
    np.savetxt(filename, save_data)


def savefig(
        filename: str, xs: np.ndarray, ys: np.ndarray, xlabel: str, ylabel: str, title=None, xlim=None, ylim=None,
        legend=None, legend_loc='upper left',figsize=(17.78, 10), dpi=400
):
    fig = plt.figure(figsize=figsize)
    for i in range(0, len(xs)):
        plt.plot(xs[i], ys[i], linewidth=0.5)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)
    if title is not None:
        plt.title(title)
    if legend is not None:
        plt.legend(legend, loc=legend_loc)
    plt.tight_layout()
    plt.savefig(filename, dpi=dpi)
    plt.close(fig)







"""
First edition on May 01 2021.
@author Jeppe Sinkbæk Thomsen, Master's student in astronomy at Aarhus University.
Supervisor: Assistant Professor Karsten Frank Brogaard.

A rotational broadening function fitting routine, with the model profile used, and all the relevant steps until the
lmfit minimizer method is called. Code is adapted from the shazam.py library for the SONG telescope
(written by Emil Knudstrup and Jens Jessen-Hansen)
"""

import numpy as np
from scipy.signal import fftconvolve
import lmfit
import scipy.constants as scc
from sb2sep.storage_classes import FitParameters


def rotational_broadening_function_profile(
        velocities: np.ndarray, amplitude: float, radial_velocity_cm: float, vsini: float, gaussian_width: float,
        continuum_constant: float, limbd_coef: float
):
    """
    Calculates a theoretical broadening function profile based on the one described in
    Kaluzny 2006: Eclipsing Binaries in the Open Cluster NGC 2243 II. Absolute Properties of NV CMa.
    Convolves it with a gaussian function to create a rotational broadening function profile.
    :param velocities:          np.ndarray, velocities to calculate profile for.
    :param amplitude:           float, normalization constant
    :param radial_velocity_cm:  float, radial velocity of the centre of mass of the star
    :param vsini:               float, linear velocity of the equator of the rotating star times sin(inclination)
    :param gaussian_width:      float, width of the gaussian broadening function that the profile will be folded with
    :param continuum_constant:  float, the continuum level
    :param limbd_coef:          float, linear limb darkening coefficient of the star.
    :return rot_bf_profile:     np.ndarray, the profile at the given velocities
    """
    n = velocities.size
    broadening_function_values = np.ones(n) * continuum_constant

    # The "a" coefficient of the profile
    a = (velocities - radial_velocity_cm) / vsini

    # Create bf function values
    mask = (np.abs(a) < 1.0)        # Select only near-peak values
    broadening_function_values[mask] += amplitude*((1-limbd_coef)*np.sqrt(1.0-a[mask]**2) + 0.25*np.pi*(1-a[mask]**2))

    # Create gs function values
    scaled_width = np.sqrt(2*np.pi) * gaussian_width
    gaussian_function_values = np.exp(-0.5 * (velocities/gaussian_width)**2) / scaled_width

    # Convolve rotational broadening function profile to get smoothed version
    rot_bf_profile = fftconvolve(broadening_function_values, gaussian_function_values, mode='same')
    return rot_bf_profile


def weight_function(
        velocities: np.ndarray, broadening_function_values: np.ndarray, velocity_fit_half_width: float,
        radial_velocity_guess: float
):
    """
    Weight function for the fit. Finds the peak value, and limits the data set to velocities within a certain distance
    of it.
    :param velocities:                  np.ndarray, the velocities of the broadening function
    :param broadening_function_values:  np.ndarray, the broadening function values
    :param velocity_fit_half_width:     float, the distance to each side within to include data for fit.
    :param radial_velocity_guess:       float, guess for the radial velocity. Used to limit weight function
    :return weight_function_values:     np.ndarray, set of 1s and 0s to weigh each data point with
    """
    if radial_velocity_guess is None:
        peak_idx = np.argmax(broadening_function_values)
        mask = (velocities > velocities[peak_idx] - velocity_fit_half_width) & \
               (velocities < velocities[peak_idx] + velocity_fit_half_width)
    else:
        mask = (velocities > radial_velocity_guess - velocity_fit_half_width) & \
               (velocities < radial_velocity_guess + velocity_fit_half_width)

    weight_function_values = np.zeros(broadening_function_values.size)
    weight_function_values[mask] = 1.0
    return weight_function_values


def get_fit_parameter_values(parameters: lmfit.Parameters):
    """
    Convenience function to pull parameter values from an lmfit.Parameters object.
    """
    amplitude = parameters['amplitude'].value
    radial_velocity_cm = parameters['radial_velocity_cm'].value
    vsini = parameters['vsini'].value
    gaussian_width = parameters['gaussian_width'].value
    continuum_constant = parameters['continuum_constant'].value
    limbd_coef = parameters['limbd_coef'].value

    return amplitude, radial_velocity_cm, vsini, gaussian_width, continuum_constant, limbd_coef


def compare_broadening_function_with_profile(
        parameters: lmfit.Parameters, velocities: np.ndarray, broadening_function_values: np.ndarray,
        weight_function_values: np.ndarray
):
    """
    Evaluates fit using rotational broadening function profile and fit parameters, and compares with the observed
    SVD broadening function. Returns the difference to the lmfit minimizer routine with weight function applied.
    :param parameters:                  lmfit.Parameters object of fit parameters
    :param velocities:                  np.ndarray, velocity values of the broadening function
    :param broadening_function_values:  np.ndarray, broadening function values
    :param weight_function_values:      np.ndarray, weight function values to apply. See weight_function() for details
    """
    parameter_vals = get_fit_parameter_values(parameters)

    comparison = broadening_function_values - rotational_broadening_function_profile(velocities, *parameter_vals)
    return weight_function_values * comparison


def fitting_routine_rotational_broadening_profile(
        velocities: np.ndarray, broadening_function_values: np.ndarray, fitparams: FitParameters,
        smooth_sigma: float, dv: float, print_report=False, compare_func=compare_broadening_function_with_profile
):
    """
    The fitting routine utilizing lmfit. Sets up the initial guesses to the fit, adds parameters, creates weight
    function, and calls the lmfit.minimize routine to fit a rotational broadening profile to match the observed
    broadening function.
    :param velocities:                  np.ndarray, velocity values of the broadening function.
    :param broadening_function_values:  np.ndarray, observed broadening function values.
    :param fitparams:                  an object holding the initial fit parameters:
            vsini_guess:                float, guess for the v sin(i) fit parameter for the model.
            limbd_coef:                 float, a linear limb darkening coefficient for the profile. This routine will
                                        not fit this parameter.
            velocity_fit_width:         float, how far out the fitting routine should include data points for the fit.
            spectral_resolution:        float, the resolution of the spectrograph used for the program spectrum.
    :param smooth_sigma:                float, the sigma value used for smoothing the broadening function values using a
                                        gaussian function.
    :param dv:                          float, the dv resolution of the interpolated spectrum in velocity space.
    :param print_report:                bool, whether to print the lmfit report after fitting. Default is False.
    :param compare_func:                function, the function used to compare the fit with the broadening function
                                        values.

    :return (fit, model):               fit: lmfit.MinimizerResult of the fit result.
                                        model: np.ndarray, model values of the broadening function according to the fit.
    """
    speed_of_light = scc.c / 1000  # in km/s
    gaussian_width = np.sqrt(((speed_of_light/fitparams.spectral_resolution)/(2.354 * dv))**2 + (smooth_sigma/dv)**2)
    params = lmfit.Parameters()

    weight_function_values = weight_function(
        velocities, broadening_function_values, fitparams.velocity_fit_width, fitparams.RV
    )
    peak_idx = np.argmax(broadening_function_values*weight_function_values)
    params.add('amplitude', value=broadening_function_values[peak_idx], min=0.0)
    if fitparams.RV is None:
        params.add(
            'radial_velocity_cm', value=velocities[peak_idx],
            min=velocities[peak_idx]-fitparams.velocity_fit_width,
            max=velocities[peak_idx]+fitparams.velocity_fit_width
        )
    else:
        params.add(
            'radial_velocity_cm', value=fitparams.RV, min=fitparams.RV-fitparams.velocity_fit_width,
            max=fitparams.RV+fitparams.velocity_fit_width
        )
    if fitparams.vsini_vary_limit is not None:
        params.add(
            'vsini', value=fitparams.vsini, vary=fitparams.vary_vsini,
            min=fitparams.vsini - fitparams.vsini*fitparams.vsini_vary_limit,
            max=fitparams.vsini + fitparams.vsini*fitparams.vsini_vary_limit
        )
    else:
        params.add('vsini', value=fitparams.vsini, vary=fitparams.vary_vsini)
    params.add('gaussian_width', value=gaussian_width, vary=False)
    params.add(
        'continuum_constant', value=fitparams.continuum, min=np.min(broadening_function_values),
        max=np.max(broadening_function_values), vary=fitparams.vary_continuum
    )
    params.add('limbd_coef', value=fitparams.limbd_coef, vary=fitparams.vary_limbd_coef)

    fit = lmfit.minimize(
        compare_func, params, args=(velocities, broadening_function_values, weight_function_values),
        xtol=1E-8, ftol=1E-8, max_nfev=500
    )
    if print_report:
        print(lmfit.fit_report(fit, show_correl=False))

    parameter_vals = get_fit_parameter_values(fit.params)
    model = rotational_broadening_function_profile(velocities, *parameter_vals)

    return fit, model



import numpy as np


def load_configuration_files(loc_routine_file, loc_separation_file, loc_rv_file):
    # Load routine config
    col0, col1 = np.genfromtxt(loc_routine_file, dtype=str, usecols=(0, 1), unpack=True)
    routine_options = RoutineOptions()
    for index, value in enumerate(col0):
        if value == 'convergence_limit':
            routine_options.convergence_limit = float(col1[index])
        elif value == 'iteration_limit':
            routine_options.iteration_limit = int(col1[index])
        elif value == 'plot':
            routine_options.plot = col1[index] == 'True'        # False if col1[index] != 'True'
        elif value == 'verbose':
            routine_options.verbose = col1[index] == 'True'
        elif value == 'save_plot_path':
            if col1[index] == 'None':
                routine_options.save_plot_path = None
            else:
                routine_options.save_plot_path = col1[index]
        elif value == 'save_all_results':
            routine_options.save_all_results = col1[index] == 'True'
        elif value == 'save_path':
            routine_options.save_path = col1[index]
        elif value == 'return_unbuffered':
            routine_options.return_unbuffered = col1[index] == 'True'
        elif value == 'plot_order':
            routine_options.plot_order = int(col1[index])
        elif value == 'adjust_vsini':
            routine_options.adjust_vsini = col1[index] == 'True'
        elif value == 'delta_v':
            routine_options.delta_v = float(col1[index])
        elif value == 'filename_bulk':
            if col1[index] == 'None':
                routine_options.filename_bulk = None
            else:
                routine_options.filename_bulk = col1[index]
        else:
            raise KeyError(f'Routine options config file key {value} not supported.')

    # Load separated spectra subroutine config file
    col0, col1 = np.loadtxt(loc_separation_file, dtype=str, usecols=(0, 1), unpack=True)
    sep_options = SeparateComponentsOptions()
    for index, value in enumerate(col0):
        if value == 'delta_v':
            sep_options.delta_v = float(col1[index])
        elif value == 'convergence_limit':
            sep_options.convergence_limit = float(col1[index])
        elif value == 'max_iterations':
            sep_options.max_iterations = int(col1[index])
        elif value == 'rv_proximity_limit':
            sep_options.rv_proximity_limit = float(col1[index])
        elif value == 'rv_lower_limit':
            sep_options.rv_lower_limit = float(col1[index])
        elif value == 'use_for_spectral_separation_A':
            if col1[index] == 'None':
                sep_options.use_for_spectral_separation_A = None
            else:
                eval_list = eval(col1[index])
                sep_options.use_for_spectral_separation_A = np.array(eval_list)
        elif value == 'use_for_spectral_separation_B':
            if col1[index] == 'None':
                sep_options.use_for_spectral_separation_B = None
            else:
                eval_list = eval(col1[index])
                sep_options.use_for_spectral_separation_B = np.array(eval_list)
        elif value == 'weights':
            if col1[index] == 'None':
                sep_options.weights = None
            else:
                eval_list = eval(col1[index])
                sep_options.weights = np.array(eval_list)
        elif value == 'verbose':
            sep_options.verbose = col1[index] == 'True'
        else:
            raise KeyError(f'Separation options config file key {value} not supported.')

    # Load RV subroutine config
    col0, col1 = np.loadtxt(loc_rv_file, dtype=str, usecols=(0, 1), unpack=True)
    rv_options = RadialVelocityOptions()
    for index, value in enumerate(col0):
        if value == 'vsini_guess_A':
            rv_options.vsini_A = float(col1[index])
        elif value == 'vsini_guess_B':
            rv_options.vsini_B = float(col1[index])
        elif value == 'vary_vsini_A':
            rv_options.vary_vsini_A = col1[index] == 'True'
        elif value == 'vary_vsini_B':
            rv_options.vary_vsini_B = col1[index] == 'True'
        elif value == 'vsini_vary_limit_A':
            if col1[index] == 'None':
                rv_options.vsini_vary_limit_A = None
            else:
                rv_options.vsini_vary_limit_A = float(col1[index])
        elif value == 'vsini_vary_limit_B':
            if col1[index] == 'None':
                rv_options.vsini_vary_limit_B = None
            else:
                rv_options.vsini_vary_limit_B = float(col1[index])
        elif value == 'delta_v':
            rv_options.delta_v = float(col1[index])
        elif value == 'spectral_resolution':
            rv_options.spectral_resolution = float(col1[index])
        elif value == 'smooth_sigma_A':
            rv_options.bf_smooth_sigma_A = float(col1[index])
        elif value == 'smooth_sigma_B':
            rv_options.bf_smooth_sigma_B = float(col1[index])
        elif value == 'bf_velocity_span':
            rv_options.bf_velocity_span = float(col1[index])
        elif value == 'velocity_fit_width_A':
            rv_options.velocity_fit_width_A = float(col1[index])
        elif value == 'velocity_fit_width_B':
            rv_options.velocity_fit_width_B = float(col1[index])
        elif value == 'refit_width_A':
            if col1[index] == 'None':
                rv_options.refit_width_A = None
            else:
                rv_options.refit_width_A = float(col1[index])
        elif value == 'refit_width_B':
            if col1[index] == 'None':
                rv_options.refit_width_B = None
            else:
                rv_options.refit_width_B = float(col1[index])
        elif value == 'limbd_coef_A':
            rv_options.limbd_coef_A = float(col1[index])
        elif value == 'limbd_coef_B':
            rv_options.limbd_coef_B = float(col1[index])
        elif value == 'vary_limbd_coef_A':
            rv_options.vary_limbd_coef_A = col1[index] == 'True'
        elif value == 'vary_limbd_coef_B':
            rv_options.vary_limbd_coef_B = col1[index] == 'True'
        elif value == 'ignore_at_phase_A':
            if col1[index] == 'None':
                rv_options.ignore_at_phase_A = None
            else:
                rv_options.ignore_at_phase_A = eval(col1[index])
        elif value == 'ignore_at_phase_B':
            if col1[index] == 'None':
                rv_options.ignore_at_phase_B = None
            else:
                rv_options.ignore_at_phase_B = eval(col1[index])
        elif value == 'verbose':
            rv_options.verbose = col1[index] == 'True'
        elif value == 'convergence_limit':
            rv_options.convergence_limit = float(col1[index])
        elif value == 'iteration_limit':
            rv_options.iteration_limit = int(col1[index])
        elif value == 'rv_lower_limit':
            rv_options.rv_lower_limit = float(col1[index])
        elif value == 'evaluate_spectra_A':
            if col1[index] == 'None':
                rv_options.evaluate_spectra_A = None
            else:
                rv_options.evaluate_spectra_A = eval(col1[index])
        elif value == 'evaluate_spectra_B':
            if col1[index] == 'None':
                rv_options.evaluate_spectra_B = None
            else:
                rv_options.evaluate_spectra_B = eval(col1[index])
        elif value == 'fitting_profile':
            if col1[index] == 'RotBF' or col1[index] == 'Gaussian':
                rv_options.fitting_profile = col1[index]
            else:
                raise ValueError("RV options: Unrecognised fitting profile selected. Either use 'RotBF' or 'Gaussian'.")
        else:
            raise KeyError(f'RV options config file key {value} not supported.')
    return routine_options, sep_options, rv_options


class FitParameters:
    def __init__(
            self,
            vsini_guess=1.0,
            spectral_resolution=60000,
            velocity_fit_width=300,
            limbd_coef=0.68,
            smooth_sigma=4.0,
            bf_velocity_span=200,
            vary_vsini=False,
            vsini_vary_limit=None,
            vary_limbd_coef=False,
            RV=None,
            continuum=0.0,
            vary_continuum=True,
            fitting_profile='RotBF'
    ):
        # Value for vsini, and whether or not to fit it
        self.vsini = vsini_guess
        self.vary_vsini = vary_vsini
        # Maximum change to vsini allowed each iteration in spectral_separation_routine()
        self.vsini_vary_limit = vsini_vary_limit
        # Data resolution
        self.spectral_resolution = spectral_resolution
        # How far away to include data in fitting procedure (rotational broadening function profile also masks
        # profile separately using (velocity - rv)/vsini )
        self.velocity_fit_width = velocity_fit_width
        # Linear limb darkening coefficient for rotational broadening function profile fit
        self.limbd_coef = limbd_coef
        self.vary_limbd_coef = vary_limbd_coef
        # Current RV values (used to update fit RV parameter limits correctly)
        self.RV = RV
        # Smoothing value (in km/s) of the convolved gaussian used in broadening function SVD (bf_smooth()).
        self.bf_smooth_sigma = smooth_sigma
        # Width of the broadening function (in velocity space)
        self.bf_velocity_span = bf_velocity_span
        # Continuum value, and whether to fit for it
        self.continuum = continuum
        self.vary_continuum = vary_continuum
        # Fitting profile to use (either 'RotBF' or 'Gaussian'
        self.fitting_profile = fitting_profile


class RadialVelocityOptions:
    def __init__(
            self,
            vsini_guess_A=1.0,
            vsini_guess_B=1.0,
            vary_vsini_A=True,
            vary_vsini_B=True,
            vsini_vary_limit_A=None,
            vsini_vary_limit_B=None,
            delta_v=1.0,
            spectral_resolution=60000,
            velocity_fit_width_A=50,
            velocity_fit_width_B=20,
            refit_width_A=None,
            refit_width_B=None,
            limbd_coef_A=0.68,
            limbd_coef_B=0.68,
            vary_limbd_coef_A=False,
            vary_limbd_coef_B=False,
            RV_A=None,
            RV_B=None,
            smooth_sigma_A=4.0,
            smooth_sigma_B=4.0,
            bf_velocity_span=200,
            ignore_at_phase_A=None,
            ignore_at_phase_B=None,
            verbose=False,
            iteration_limit=6,
            convergence_limit=5e-3,
            rv_lower_limit=0.0,
            print_prec=6,
            n_parallel_jobs=1,
            evaluate_spectra_A=None,
            evaluate_spectra_B=None,
            fitting_profile='RotBF'
    ):
        # Value for vsini, and whether or not to fit it
        self.vsini_A = vsini_guess_A
        self.vsini_B = vsini_guess_B
        self.vary_vsini_A = vary_vsini_A
        self.vary_vsini_B = vary_vsini_B

        # Maximum change to vsini allowed each iteration in spectral_separation_routine()
        self.vsini_vary_limit_A = vsini_vary_limit_A
        self.vsini_vary_limit_B = vsini_vary_limit_B

        # Data resolution
        self.spectral_resolution = spectral_resolution
        self.delta_v = delta_v

        # How far away to include data in fitting procedure (rotational broadening function profile also masks
        # profile separately using (velocity - rv)/vsini )
        self.velocity_fit_width_A = velocity_fit_width_A
        self.velocity_fit_width_B = velocity_fit_width_B
        # Repeat and refit results with different width after honing in?
        self.refit_width_A = refit_width_A
        self.refit_width_B = refit_width_B

        # Linear limb darkening coefficient for rotational broadening function profile fit
        self.limbd_coef_A = limbd_coef_A
        self.limbd_coef_B = limbd_coef_B
        self.vary_limbd_coef_A = vary_limbd_coef_A
        self.vary_limbd_coef_B = vary_limbd_coef_B

        # Current RV values (used to update fit RV parameter limits correctly)
        self.RV_A = RV_A
        self.RV_B = RV_B

        # Smoothing value (in km/s) of the convolved gaussian used in broadening function SVD (bf_smooth()).
        self.bf_smooth_sigma_A = smooth_sigma_A
        self.bf_smooth_sigma_B = smooth_sigma_B

        # Width of the broadening function (in velocity space)
        self.bf_velocity_span = bf_velocity_span

        # Use if component should not be subtracted in a specific phase-area (fx. (0.7, 0.9)), if it is totally eclipsed
        self.ignore_at_phase_A = ignore_at_phase_A
        self.ignore_at_phase_B = ignore_at_phase_B

        # Can be 'False', 'True' (or 'all') or 'errors'
        self.verbose=verbose

        self.iteration_limit=iteration_limit
        self.convergence_limit=convergence_limit

        self.rv_lower_limit = rv_lower_limit

        self.print_prec = print_prec

        self.n_parallel_jobs = n_parallel_jobs

        self.evaluate_spectra_A = evaluate_spectra_A
        self.evaluate_spectra_B = evaluate_spectra_B

        self.fitting_profile = fitting_profile


class SeparateComponentsOptions:
    def __init__(
            self,
            delta_v=1.0,
            convergence_limit=1e-2,
            max_iterations=10,
            rv_proximity_limit=0.0,
            rv_lower_limit=0.0,
            ignore_component_B=False,
            verbose=False,
            use_for_spectral_separation_A=None,
            use_for_spectral_separation_B=None,
            weights=None
    ):
        self.delta_v = delta_v
        self.convergence_limit = convergence_limit
        self.max_iterations = max_iterations
        self.rv_proximity_limit = rv_proximity_limit
        self.rv_lower_limit = rv_lower_limit
        self.ignore_component_B = ignore_component_B
        self.verbose = verbose
        self.use_for_spectral_separation_A = use_for_spectral_separation_A
        self.use_for_spectral_separation_B = use_for_spectral_separation_B
        self.weights = weights


class RoutineOptions:
    def __init__(
            self,
            convergence_limit=1E-5,
            iteration_limit=10,
            plot=True,
            verbose=True,
            return_unbuffered=True,
            save_plot_path=None,
            save_all_results=True,
            save_path='./',
            filename_bulk=None,
            plot_order=0,
            adjust_vsini=True,
            delta_v=1.0
    ):
        self.convergence_limit = convergence_limit
        self.iteration_limit = iteration_limit
        self.plot = plot
        self.verbose = verbose
        self.return_unbuffered = return_unbuffered
        self.save_plot_path = save_plot_path
        self.save_all_results = save_all_results
        self.save_path = save_path
        self.filename_bulk = filename_bulk
        self.plot_order = plot_order
        self.adjust_vsini = adjust_vsini
        self.delta_v = delta_v



from sb2sep.broadening_function_svd import *
import sb2sep.rotational_broadening_function_fitting as rbff
import sb2sep.gaussian_fitting as gf
from copy import copy
from joblib import Parallel, delayed
from sb2sep.storage_classes import FitParameters, RadialVelocityOptions
import numpy as np


def radial_velocity_2_components(
        inv_flux, broadening_function_template:BroadeningFunction,
        options: RadialVelocityOptions
):
    BFsvd = copy(broadening_function_template)
    BFsvd.spectrum = inv_flux
    BFsvd.smooth_sigma = options.bf_smooth_sigma_A

    # Create Broadening Function for star A
    BFsvd.solve()
    BFsvd.smooth()

    # Fit rotational broadening function profile to Giant peak
    fitparams_A = FitParameters(
        options.vsini_A, options.spectral_resolution, options.velocity_fit_width_A, options.limbd_coef_A,
        options.bf_smooth_sigma_A, options.bf_velocity_span, options.vary_vsini_A, options.vsini_vary_limit_A,
        options.vary_limbd_coef_A, options.RV_A
    )
    fit_A, model_values_A = BFsvd.fit_rotational_profile(fitparams_A)

    # Create Broadening Function for star B
    bf, bf_smooth = BFsvd.bf, BFsvd.bf_smooth
    BFsvd.smooth_sigma = options.bf_smooth_sigma_B
    BFsvd.bf = BFsvd.bf - model_values_A        # subtract model for giant
    BFsvd.smooth()

    # Fit rotational broadening function profile for MS peak
    if options.limbd_coef_B is None:
        options.limbd_coef_B = options.limbd_coef_A
    fitparams_B = FitParameters(
        options.vsini_B, options.spectral_resolution, options.velocity_fit_width_B, options.limbd_coef_B,
        options.bf_smooth_sigma_B, options.bf_velocity_span, options.vary_vsini_B, options.vsini_vary_limit_B,
        options.vary_limbd_coef_B, options.RV_B
    )
    fit_B, model_values_B = BFsvd.fit_rotational_profile(fitparams_B)

    _, RV_A, _, _, _, _ = get_fit_parameter_values(fit_A.params)
    _, RV_B, _, _, _, _ = get_fit_parameter_values(fit_B.params)
    return (RV_A, RV_B), (model_values_A, fit_A, model_values_B, fit_B), (bf, bf_smooth)


def _pull_results_mspectra_2comp(broadening_function_template: BroadeningFunction, res_par, n_spectra, plot):
    RVs_A = np.empty((n_spectra,))
    RVs_B = np.empty((n_spectra,))
    broadening_function_vals = np.empty((broadening_function_template.velocity.size, n_spectra))
    broadening_function_vals_smoothed = np.empty((broadening_function_template.velocity.size, n_spectra))
    model_values_A = np.empty((broadening_function_template.velocity.size, n_spectra))
    model_values_B = np.empty((broadening_function_template.velocity.size, n_spectra))
    for i in range(0, n_spectra):
        RV_values = res_par[i][0]
        models, (bf, bf_smooth) = res_par[i][1], res_par[i][2]
        broadening_function_vals[:, i] = bf
        broadening_function_vals_smoothed[:, i] = bf_smooth
        model_values_A[:, i] = models[0]
        model_values_B[:, i] = models[2]
        if plot:
            plt.figure()
            plt.plot(broadening_function_template.velocity, bf_smooth)
            plt.plot(broadening_function_template.velocity, models[0], 'k--')
            plt.plot(broadening_function_template.velocity, models[2], 'k--')
            plt.show(block=False)
        RVs_A[i], RVs_B[i] = RV_values[0], RV_values[1]
    if plot:
        plt.show(block=True)
    extra_results = (broadening_function_template.velocity, broadening_function_vals, broadening_function_vals_smoothed,
                     model_values_A, model_values_B)
    return RVs_A, RVs_B, extra_results


def _pull_results_mspectra_1comp(broadening_function_template: BroadeningFunction, res_par, n_spectra, plot):
    RVs = np.empty((n_spectra, ))
    broadening_function_vals = np.empty((broadening_function_template.velocity.size, n_spectra))
    broadening_function_vals_smoothed = np.empty((broadening_function_template.velocity.size, n_spectra))
    model_values = np.empty((broadening_function_template.velocity.size, n_spectra))
    for i in range(0, n_spectra):
        RVs[i] = res_par[i][0]
        fit_res = res_par[i][1]
        model_values[:, i], broadening_function_vals[:, i] = fit_res[1], fit_res[3]
        broadening_function_vals_smoothed[:, i] = fit_res[4]
        if plot:
            plt.figure()
            plt.plot(broadening_function_template.velocity, broadening_function_vals_smoothed[:, i])
            plt.plot(broadening_function_template.velocity, model_values, 'k--')
            plt.show(block=False)
    if plot:
        plt.show(block=True)
    extra_results = (broadening_function_template.velocity, broadening_function_vals, broadening_function_vals_smoothed,
                     model_values)
    return RVs, extra_results


def radial_velocities_of_multiple_spectra(
        inv_flux_collection: np.ndarray,
        inv_flux_template: np.ndarray,
        options: RadialVelocityOptions,
        fit_two_components=True,
        number_of_parallel_jobs=4,
        plot=False
):
    """
    Calculates radial velocities for two components, for multiple spectra, by fitting two rotational broadening function
    profiles successively to calculated broadening functions. Uses joblib to parallelize the calculations. Calls
    radial_velocity_from_broadening_function() for each spectrum.

    Assumes that both components are well-described by the single template spectrum provided for the broadening function
    calculation.

    :param inv_flux_collection:     np.ndarray shape (:, n_spectra). Collection of inverted fluxes for each spectrum.
    :param inv_flux_template:       np.ndarray shape (:, ). Inverted flux of a template spectrum.
    :param options:                 fit options passed on from above.
    :param fit_two_components:      bool. Indicates if 1 or 2 components should be fitted for.
    :param number_of_parallel_jobs: int. Indicates the number of separate processes to spawn with joblib.
    :param plot:                    bool. Indicates whether results should be plotted, with separate figures for each
                                        spectrum.
    :return:    RVs_A, RVs_B, extra_results.
                RVs_A:  np.ndarray shape (n_spectra, ). RV values of component A.
                RVs_B:  np.ndarray shape (n_spectra, ). RV values of component B.
                extra_results: (broadening function velocity values, broadening function values,
                                smoothed broadening function values, fit model values for component A,
                                fit model values for component B).
    """
    n_spectra = inv_flux_collection[0, :].size
    broadening_function_template = BroadeningFunction(inv_flux_collection[:, 0], inv_flux_template,
                                                      options.bf_velocity_span, options.delta_v)
    broadening_function_template.smooth_sigma = options.bf_smooth_sigma_A

    # Arguments for parallel job
    if fit_two_components is True:
        arguments = (broadening_function_template, options)
        calc_function = radial_velocity_2_components
    else:
        fitparams = FitParameters(
            options.vsini_A, options.spectral_resolution, options.velocity_fit_width_A, options.limbd_coef_A,
            options.bf_smooth_sigma_A, options.bf_velocity_span, options.vary_vsini_A, options.vsini_vary_limit_A,
            options.vary_limbd_coef_A, options.RV_A
        )
        arguments = (broadening_function_template, fitparams)
        calc_function = radial_velocity_single_component

    # Create parallel call to calculate radial velocities
    res_par = Parallel(n_jobs=number_of_parallel_jobs)(
        delayed(calc_function)(inv_flux_collection[:, i], *arguments) for i in range(0, n_spectra)
    )

    # Pull results from call
    if fit_two_components is True:
        RVs_A, RVs_B, extra_results = _pull_results_mspectra_2comp(
            broadening_function_template, res_par, n_spectra, plot
        )
        bf_velocity, bf_vals, bf_vals_smooth, model_vals_A, model_vals_B = extra_results
        return RVs_A, RVs_B, (bf_velocity, bf_vals, bf_vals_smooth, model_vals_A, model_vals_B)
    else:
        RVs, (bf_velocity, bf_vals, bf_vals_smooth, model_vals) = _pull_results_mspectra_1comp(
            broadening_function_template, res_par, n_spectra, plot
        )
        return RVs, (bf_velocity, bf_vals, bf_vals_smooth, model_vals)


def radial_velocity_single_component(
        inv_flux: np.ndarray,
        broadening_function_template: BroadeningFunction,
        fitparams: FitParameters
):
    """
    Calculates the broadening function of a spectrum and fits a single rotational broadening function profile to it.
    Needs a template object of the BroadeningFunction with the correct parameters and template spectrum already set. 
    Setup to be of convenient use during the spectral separation routine (see spectral_separation_routine.py).
    
    :param inv_flux:                     np.ndarray. Inverted flux of the program spectrum (e.g. 1-normalized_flux)
    :param broadening_function_template: BroadeningFunction. The template used to calculate the broadening function.
    :param fitparams:                   InitialFitParameters. Object that stores the fitting parameters needed.
    :return:    RV, (fit, model_values, BFsvd.velocity, BFsvd.bf, BFsvd.bf_smooth)
                RV:                 float, the fitted RV value
                fit:                lmfit.MinimizerResult. The object storing the fit parameters.
                model_values:       np.ndarray. Broadening function values according to the fit.
                BFsvd.velocity:     np.ndarray. Velocity values for the broadening function and model values.
                BFsvd.bf:           np.ndarray. Broadening function values calculated.
                BFsvd.bf_smooth:    np.ndarray. Smoothed broadening function values.
    """
    BFsvd = copy(broadening_function_template)
    BFsvd.spectrum = inv_flux

    # Create Broadening Function
    BFsvd.solve()
    BFsvd.smooth()

    # Fit rotational broadening function profile
    if fitparams.fitting_routine == 'RotBF':
        fitting_routine = rbff.fitting_routine_rotational_broadening_profile
        get_fit_parameter_values = rbff.get_fit_parameter_values
    elif fitparams.fitting_routine == 'Gaussian':
        fitting_routine = gf.fitting_routine_gaussian_profile
        get_fit_parameter_values = gf.get_fit_parameter_values
    else:
        raise ValueError('Unrecognised fitting profile selected.')
    fit, model_values = BFsvd.fit_rotational_profile(fitparams, fitting_routine)

    res = get_fit_parameter_values(fit.params)
    RV = res[1]

    return RV, (fit, model_values, BFsvd.velocity, BFsvd.bf, BFsvd.bf_smooth)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib


"""
This file is meant to be used with the additional data results provided by the spectral separation routine
(can be turned on by using the save_additional_results switch in spectral_separation_routine()). Default location for
these files are in RV/Data/additionals/spectral_separation/.

Two storage classes are provided, IntervalResult (where loaded data is stored) and RoutineResults (where multiple
IntervalResult objects are accessed). Datafiles from the separation routine can be loaded into these objects using 
load_routine_results(), which provides a RoutineResults instance that can be fed into the plotting functions.
load_routine_results() requests a folder path, as well as a list of "bulk" filenames, e.g. the wavelength-intervals. An
example would be:
load_routine_results(RV/Data/additionals/spectral_separation/, ['4500_5000', '5000_5500'])
"""


class IntervalResult:
    def __init__(
            self, time_values_A, time_values_B,
            wavelength, separated_spectrum_A, separated_spectrum_B, template_spectrum_A, template_spectrum_B,
            RV_A, RV_B, RV_A_initial, RV_B_initial,
            bf_velocity_A, bf_vals_A, bf_smooth_A, bf_model_vals_A,
            bf_velocity_B, bf_vals_B, bf_smooth_B, bf_model_vals_B,
            wavelength_a=None, wavelength_b=None
    ):
        self.time_values_A = time_values_A
        self.time_values_B = time_values_B
        self.wavelength = wavelength
        self.separated_spectrum_A = separated_spectrum_A
        self.separated_spectrum_B = separated_spectrum_B
        self.template_flux_A = template_spectrum_A
        self.template_flux_B = template_spectrum_B
        self.RV_A = RV_A
        self.RV_B = RV_B
        self.RV_A_initial = RV_A_initial
        self.RV_B_initial = RV_B_initial
        self.bf_velocity_A = bf_velocity_A
        self.bf_vals_A = bf_vals_A
        self.bf_smooth_A = bf_smooth_A
        self.bf_model_vals_A = bf_model_vals_A
        self.bf_velocity_B = bf_velocity_B
        self.bf_vals_B = bf_vals_B
        self.bf_smooth_B = bf_smooth_B
        self.bf_model_vals_B = bf_model_vals_B
        if wavelength_a is None:
            self.wavelength_a = int(np.round(np.min(wavelength)))
        if wavelength_b is None:
            self.wavelength_b = int(np.round(np.max(wavelength)))


class RoutineResults:
    def __init__(self, *args: IntervalResult):
        if len(args) != 0:
            self.interval_results = [x for x in args]
        else:
            self.interval_results = []

    @property
    def time_values_A(self):
        return [x.time_values_A for x in self.interval_results]

    @property
    def time_values_B(self):
        return [x.time_values_B for x in self.interval_results]

    @property
    def RV_A(self):
        return [x.RV_A for x in self.interval_results]

    @property
    def RV_B(self):
        return [x.RV_B for x in self.interval_results]

    @property
    def RV_A_initial(self):
        return [x.RV_A_initial for x in self.interval_results]

    @property
    def RV_B_initial(self):
        return [x.RV_B_initial for x in self.interval_results]

    @property
    def wavelengths(self, interval=None):
        if interval is not None:
            res = None
            for interval_result in self.interval_results:
                if interval_result.wavelength_a == interval[0] and interval_result.wavelength_b == interval[1]:
                    return interval_result.wavelength
            if res is None:
                raise ValueError('No interval found with the same wavelength_a and wavelength_B.')
        else:
            return [x.wavelength for x in self.interval_results]

    @property
    def wavelength_a(self):
        return [x.wavelength_a for x in self.interval_results]

    @property
    def wavelength_b(self):
        return [x.wavelength_b for x in self.interval_results]

    @property
    def interval(self):
        return [(x.wavelength_a, x.wavelength_b) for x in self.interval_results]

    @property
    def separated_spectra_A(self):
        return [x.separated_spectrum_A for x in self.interval_results]

    @property
    def separated_spectra_B(self):
        return [x.separated_spectrum_B for x in self.interval_results]

    @property
    def template_flux_A(self):
        return [x.template_flux_A for x in self.interval_results]

    @property
    def template_flux_B(self):
        return [x.template_flux_B for x in self.interval_results]

    @property
    def bf_results(self):
        return [
            [x.bf_velocity_A, x.bf_vals_A, x.bf_smooth_A, x.bf_model_vals_A, x.bf_velocity_B, x.bf_vals_B,
             x.bf_smooth_B, x.bf_model_vals_B]
            for x in self.interval_results
        ]

    def append_interval(self, new_result: IntervalResult):
        self.interval_results.append(new_result)


def load_routine_results(folder_path: str, filename_bulk_list: list):
    routine_results = RoutineResults()
    for filename_bulk in filename_bulk_list:
        try:
            rvA_array = np.loadtxt(folder_path+filename_bulk+'_rvA.txt')
            rvB_array = np.loadtxt(folder_path+filename_bulk+'_rvB.txt')
        except OSError:
            rvA_array = np.loadtxt(folder_path + '_rvA.txt')
            rvB_array = np.loadtxt(folder_path + '_rvB.txt')
        sep_array = np.loadtxt(folder_path+filename_bulk+'_sep_flux.txt')
        velA_array = np.loadtxt(folder_path+filename_bulk+'_velocities_A.txt')
        bfA_array = np.loadtxt(folder_path+filename_bulk+'_bfvals_A.txt')
        bfA_smooth_array = np.loadtxt(folder_path+filename_bulk+'_bfsmooth_A.txt')
        modelA_array = np.loadtxt(folder_path+filename_bulk+'_models_A.txt')
        velB_array = np.loadtxt(folder_path+filename_bulk+'_velocities_B.txt')
        bfB_array = np.loadtxt(folder_path+filename_bulk+'_bfvals_B.txt')
        bfB_smooth_array = np.loadtxt(folder_path+filename_bulk+'_bfsmooth_B.txt')
        modelB_array = np.loadtxt(folder_path+filename_bulk+'_models_B.txt')
        rvs_initial = np.loadtxt(folder_path+filename_bulk+'_rv_initial.txt')

        wavelength, separated_flux_A, separated_flux_B = sep_array[:, 0], sep_array[:, 1], sep_array[:, 2]
        template_flux_A, template_flux_B = sep_array[:, 3], sep_array[:, 4]
        if rvA_array[0, :].size != 2:
            time_values_A, RV_A = rvA_array[:, 0], rvA_array[:, 1:]
            time_values_B, RV_B = rvB_array[:, 0], rvB_array[:, 1:]
        else:
            time_values_A, RV_A = rvA_array[:, 0], rvA_array[:, 1]
            time_values_B, RV_B = rvB_array[:, 0], rvB_array[:, 1]
        RV_A_initial, RV_B_initial = rvs_initial[:, 0], rvs_initial[:, 1]

        interval_result = IntervalResult(
            time_values_A, time_values_B, wavelength, separated_flux_A, separated_flux_B, template_flux_A,
            template_flux_B, RV_A, RV_B, RV_A_initial, RV_B_initial, velA_array, bfA_array,
            bfA_smooth_array, modelA_array, velB_array, bfB_array, bfB_smooth_array, modelB_array
        )
        routine_results.append_interval(interval_result)
    return routine_results


def plot_rv_and_separated_spectra(
        evaluation_data: RoutineResults,
        period: float,
        block=False,
        color_A='b',
        color_B='r',
        fig=None,
        axs: list = None
):
    matplotlib.rcParams.update({'font.size': 20})
    for i in range(0, len(evaluation_data.interval_results)):
        if fig is None:
            fig = plt.figure(figsize=(16, 9))
        if axs is None:
            gspec = fig.add_gridspec(2, 2)
            ax1 = fig.add_subplot(gspec[0, :])
            ax2 = fig.add_subplot(gspec[1, 0])
            ax3 = fig.add_subplot(gspec[1, 1])
        else:
            ax1, ax2, ax3 = axs

        phase_A = np.mod(evaluation_data.time_values_A[i], period) / period
        phase_B = np.mod(evaluation_data.time_values_B[i], period) / period

        if evaluation_data.RV_A[i].ndim == 1:
            ax1.plot(phase_A, evaluation_data.RV_A[i], color_A+'*')
            ax1.plot(phase_B, evaluation_data.RV_B[i], color_B+'*')
        else:
            ax1.plot(phase_A, np.mean(evaluation_data.RV_A[i]), color_A + '*')
            ax1.plot(phase_B, np.mean(evaluation_data.RV_B[i]), color_B + '*')
        ax1.set_xlabel('Orbital Phase')
        ax1.set_ylabel('Radial Velocity - system velocity (km/s)', fontsize=15)

        ax2.plot(evaluation_data.wavelengths[i], evaluation_data.template_flux_A[i], '--', color='grey', linewidth=0.5)
        ax2.plot(evaluation_data.wavelengths[i], evaluation_data.separated_spectra_A[i], color_A, linewidth=2)
        ax2.set_xlabel('Wavelength (Å)')
        ax2.set_ylabel('Normalized Separated Flux', fontsize=15)

        ax3.plot(evaluation_data.wavelengths[i], evaluation_data.template_flux_B[i], '--', color='grey', linewidth=0.5)
        ax3.plot(evaluation_data.wavelengths[i], evaluation_data.separated_spectra_B[i], color_B, linewidth=2)
        ax2.set_xlabel('Wavelength (Å)')
        ax3.set_xlabel('Wavelength (Å)')
        fig.suptitle(f'Interval results {evaluation_data.wavelength_a[i]}-{evaluation_data.wavelength_b[i]} Å ')
        plt.tight_layout()

    plt.show(block=block)
    return fig, ax1, ax2, ax3


def plot_broadening_functions(evaluation_data: RoutineResults, block=False, xlim=None, smoothed=True):
    matplotlib.rcParams.update({'font.size': 25})
    for i in range(0, len(evaluation_data.interval_results)):
        fig = plt.figure(figsize=(16, 15))
        gspec = fig.add_gridspec(1, 2)
        ax1 = fig.add_subplot(gspec[0, 0])
        ax2 = fig.add_subplot(gspec[0, 1])

        bf_results = evaluation_data.bf_results[i]
        if smoothed is True:
            vel_A, bf_A, models_A = bf_results[0], bf_results[2], bf_results[3]
            vel_B, bf_B, models_B = bf_results[4], bf_results[6], bf_results[7]
        else:
            vel_A, bf_A, models_A = bf_results[0], bf_results[1], bf_results[3]
            vel_B, bf_B, models_B = bf_results[4], bf_results[5], bf_results[7]
        RV_A = evaluation_data.RV_A[i]
        RV_B = evaluation_data.RV_B[i]
        RV_B_initial = evaluation_data.RV_B_initial[i]

        if xlim is None:
            ax1.set_xlim([np.min(vel_A), np.max(vel_A)])
            ax2.set_xlim([np.min(vel_A), np.max(vel_A)])
        else:
            ax1.set_xlim(xlim)
            ax2.set_xlim(xlim)

        ax1.plot(np.zeros(shape=(2,)), [0.0, 1.05], '--', color='grey')
        ax2.plot(np.zeros(shape=(2,)), [0.0, 1.05], '--', color='grey')
        for k in range(0, vel_A[:, 0].size):
            offset = k/vel_A[:, 0].size
            scale = (0.5/vel_A[:, 0].size)/np.max(bf_A[k, :])
            ax1.plot(vel_A[k, :], 1 + scale*bf_A[k, :] - offset)
            ax1.plot(vel_A[k, :], 1 + scale*models_A[k, :] - offset, 'k--')
            # ax1.plot(np.ones(shape=(2,))*RV_A[k], [1-offset*1.01, 1+5/4 * scale*np.max(models_A[k, :])-offset],
            #          color='blue')

            scale = (0.5/vel_B[:, 0].size)/np.max(bf_B[k, :])
            ax2.plot(vel_B[k, :], 1 + scale*bf_B[k, :] - offset)
            ax2.plot(vel_B[k, :], 1 + scale*models_B[k, :] - offset, 'k--')
            # ax2.plot(np.ones(shape=(2,)) * RV_B[k], [1-offset*1.01, 1+5/4 * scale*np.max(models_B[k, :])-offset],
            #          color='blue')
            # ax2.plot(np.ones(shape=(2,)) * RV_B_initial[k], [1-offset*1.01, 1+5/4*scale*np.max(models_B[k, :])-offset],
            #          color='grey')
            if smoothed is True:
                ax1.set_ylabel('Normalized, smoothed Broadening Function')
            else:
                ax1.set_ylabel('Normalized Broadening Function')
            ax1.set_xlabel('Velocity Shift [km/s]')
            ax2.set_xlabel('Velocity Shift [km/s]')
            fig.suptitle(f'Interval results {evaluation_data.wavelength_a[i]}-{evaluation_data.wavelength_b[i]} Å ')
            ax1.tick_params(
                axis='y',  # changes apply to the x-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=False)  # labels along the bottom edge are off
            ax2.tick_params(
                axis='y',  # changes apply to the x-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=False)  # labels along the bottom edge are off
            plt.tight_layout()
    plt.show(block=block)


def _create_telluric_spectrum(wavelength_a=4000.0, wavelength_b=9500.0, plot=False):
    path_telluric = 'RV/Data/template_spectra/telluric_noheader.txt'

    def gaussian(x, amp, center, fwhm_):
        return amp * np.exp(-(x-center)**2 / (2*(fwhm_/2.35482)**2))

    wavelength = np.linspace(wavelength_a, wavelength_b, 5000)
    flux = np.ones(shape=wavelength.shape)
    telluric_data = np.loadtxt(path_telluric)

    for i in range(0, telluric_data[:, 0].size):
        w1, w2 = telluric_data[i, 0], telluric_data[i, 1]
        fwhm = w2 - w1
        mask = (wavelength > w1-2*fwhm) & (wavelength < w2+2*fwhm)
        residual_intensity, line_center = telluric_data[i, 2], telluric_data[i, 3]
        if mask.size != 0:
            flux[mask] -= gaussian(wavelength[mask], 1-residual_intensity, line_center, fwhm)

    if plot:
        plt.figure()
        plt.plot(wavelength, flux)
        plt.show(block=True)

    return wavelength, flux


def compare_interval_result(evaluation_data: RoutineResults, spectrum_number: int, block=True, ax=None, hide_bad=False):
    if ax is None:
        matplotlib.rcParams.update({'font.size': 25})
        fig, ax_ = plt.subplots(figsize=(16, 9))
    else:
        ax_ = ax

    w_a = np.empty((len(evaluation_data.interval_results), ))
    w_b = np.empty((len(evaluation_data.interval_results), ))
    RV_B = np.empty((len(evaluation_data.interval_results), ))
    w_a[:] = np.nan
    w_b[:] = np.nan
    RV_B[:] = np.nan
    for i in range(0, len(evaluation_data.interval_results)):
        if hide_bad is True and evaluation_data.RV_B_flags[i][spectrum_number] == 0.0:
            pass
        else:
            w_a[i] = evaluation_data.wavelength_a[i]
            w_b[i] = evaluation_data.wavelength_b[i]
            RV_B[i] = evaluation_data.RV_B[i][spectrum_number]

    ax_.errorbar(w_a + (w_b-w_a)/2, RV_B, xerr=(w_b-w_a)/2, fmt='*')
    telluric_wl, telluric_fl = _create_telluric_spectrum(np.min(w_a), np.max(w_b))
    ax_.plot(telluric_wl, telluric_fl*30-55, '--', color='grey', linewidth=0.5)

    if ax is None:
        ax_.set_xlabel('Wavelength (Å)')
        ax_.set_ylabel('Component B RV - SV (km/s)')
        plt.show(block=block)


def compare_interval_results_multiple_spectra(evaluation_data: RoutineResults, block=True, hide_bad=True):
    matplotlib.rcParams.update({'font.size': 25})
    fig, ax = plt.subplots(figsize=(16, 9))

    for i in range(0, evaluation_data.RV_B[0].size):
        compare_interval_result(evaluation_data, spectrum_number=i, ax=ax, hide_bad=hide_bad)

    ax.set_xlabel('Wavelength (Å)')
    ax.set_ylabel('Component B RV - SV (km/s)')

    plt.show(block=block)





"""
First edition May 02/2021.
Package version: December 08/2021.
@author Jeppe Sinkbæk Thomsen, Master's thesis student at Aarhus University.
Supervisor: Karsten Frank Brogaard.

This is a collection of functions that form a routine to perform spectral separation of detached eclipsing binaries
with a giant component and a main sequence component. The routine is adapted from multiple scripts, an IDL script by
Karsten Frank Brogaard named "dis_real_merged_8430105_2021.pro", and a similar python script "spec_8430105_bf.py" by the
same author. Both follows the formula layout of the article:
'Separation of composite spectra: the spectroscopic detecton of an eclipsing binary star'
        by J.F. Gonzalez and H. Levato ( A&A 448, 283-292(2006) )
"""
import matplotlib.pyplot as plt

from sb2sep.calculate_radial_velocities import radial_velocity_single_component
from sb2sep.broadening_function_svd import *
from sb2sep.storage_classes import FitParameters, SeparateComponentsOptions, RadialVelocityOptions, \
    RoutineOptions
import sb2sep.spectrum_processing_functions as spf
import numpy as np
from copy import deepcopy
from joblib import Parallel, delayed
import matplotlib
import sb2sep.broadening_function_svd as bfsvd
# from scipy.interpolate import interp1d
import scipy.constants as scc
from matplotlib.backends.backend_pdf import PdfPages
from copy import copy


def shift_spectrum(flux, radial_velocity_shift, delta_v):
    """
    Performs a wavelength shift of a spectrum. Two-step process:
     - First, it performs a low resolution roll of the flux elements. Resolution here is defined by delta_v (which must
       be the velocity spacing of the flux array). Example precision: 1 km/s (with delta_v=1.0)
     - Second, it performs linear interpolation to produce a small shift to correct for the low precision.
     The spectrum must be evenly spaced in velocity space in order for this function to properly work.
    :param flux:                    np.ndarray size (:, ), flux values of the spectrum.
    :param radial_velocity_shift:   float, the shift in velocity units (km/s)
    :param delta_v:                 float, the resolution of the spectrum in km/s.
    :return flux_shifted:           np.ndarray size (:, ), shifted flux values of the spectrum.

    shift = radial_velocity_shift / delta_v
    indices = np.linspace(0, flux.size, flux.size)
    indices_shifted = indices - shift
    flux_shifted = np.interp(indices_shifted, indices, flux)
    """
    # Perform large delta_v precision shift
    large_shift = np.floor(radial_velocity_shift / delta_v)
    flux_ = np.roll(flux, int(large_shift))

    # Perform small < delta_v precision shift using linear interpolation
    small_shift = radial_velocity_shift / delta_v - large_shift
    indices = np.linspace(0, flux.size, flux.size)
    indices_shifted = indices - small_shift  # minus will give same shift direction as np.roll()
    # flux_shifted = interp1d(indices, flux_, fill_value='extrapolate')(indices_shifted)
    flux_shifted = np.interp(indices_shifted, indices, flux_)

    return flux_shifted


def shift_wavelength_spectrum(wavelength, flux, radial_velocity_shift):
    wavelength_shifted = wavelength * (1 + radial_velocity_shift/(scc.speed_of_light / 1000))
    # flux_shifted = interp1d(wavelength_shifted, flux, fill_value='extrapolate')(wavelength)
    flux_shifted = np.interp(wavelength, wavelength_shifted, flux)
    return flux_shifted


def separate_component_spectra(
        flux_collection, radial_velocity_collection_A, radial_velocity_collection_B,
        options: SeparateComponentsOptions, outside_separated_A=None, outside_separated_B=None,
        weights_outside_A=None, weights_outside_B=None
):
    """
    Assumes that component A is the dominant component in the spectrum. Attempts to separate the two components using
    RV shifts and averaged spectra.

    :param flux_collection:               np.ndarray shape (datasize, nspectra) of all the observed spectra
    :param radial_velocity_collection_A:  np.ndarray shape (nspectra, ) of radial velocity values for component A
    :param radial_velocity_collection_B:  np.ndarray shape (nspectra, ) of radial velocity values for component B
    :param options:                       options for the subroutine, including convergence tolerances, spectrum weights
                                            and more

    :return separated_flux_A, separated_flux_B:   the separated and meaned total component spectra of A and B.
    """
    n_spectra = flux_collection[0, :].size
    if outside_separated_A is not None and weights_outside_A is not None:
        # separated_flux_A = outside_separated_A
        separated_flux_A = np.zeros((flux_collection[:, 0].size,))
    else:
        separated_flux_A = np.zeros((flux_collection[:, 0].size,))
    if outside_separated_B is not None and weights_outside_B is not None:
        # separated_flux_B = outside_separated_B
        separated_flux_B = np.zeros((flux_collection[:, 0].size,))
    else:
        separated_flux_B = np.zeros((flux_collection[:, 0].size,))  # Set to 0 before iteration

    use_spectra_A = options.use_for_spectral_separation_A
    use_spectra_B = options.use_for_spectral_separation_B

    delta_v = options.delta_v
    weights = options.weights
    if weights is None:
        weights = np.ones(n_spectra)

    iteration_counter = 0
    while True:
        RMS_values_A = -separated_flux_A
        RMS_values_B = -separated_flux_B
        iteration_counter += 1

        if weights_outside_A is None:
            separated_flux_A[:] = 0.0
            n_used_spectra = 0
        else:
            separated_flux_A = weights_outside_A * outside_separated_A
            n_used_spectra = np.copy(weights_outside_A)
        for i in range(0, n_spectra):
            if (use_spectra_A is None) or (i in use_spectra_A):
                rvA = radial_velocity_collection_A[i]
                rvB = radial_velocity_collection_B[i]
                if options.rv_lower_limit == 0.0 and use_spectra_A is None:
                    condition = np.abs(rvA-rvB) > options.rv_proximity_limit
                elif use_spectra_A is None:
                    condition = np.abs(rvA) > options.rv_lower_limit
                else:
                    condition = True

                if condition:
                    shifted_flux_A = shift_spectrum(flux_collection[:, i], -rvA, delta_v)
                    if options.ignore_component_B is False:
                        separated_flux_A += weights[i] * \
                                            (shifted_flux_A - shift_spectrum(separated_flux_B, rvB - rvA, delta_v))
                    else:
                        separated_flux_A += weights[i] * shifted_flux_A
                    n_used_spectra += weights[i]
            elif use_spectra_A.size != 0:
                pass
            else:
                raise TypeError(f'use_spectra_A is either of wrong type ({type(use_spectra_A)}), empty, or wrong value.\n' +
                                f'Expected type: {type(True)} or np.ndarray. Expected value if bool: True')
        separated_flux_A = separated_flux_A / n_used_spectra

        if weights_outside_B is None:
            separated_flux_B[:] = 0.0
            n_used_spectra = 0
        else:
            separated_flux_B = weights_outside_B * outside_separated_B
            n_used_spectra = np.copy(weights_outside_B)
        for i in range(0, n_spectra):
            if (use_spectra_B is None) or (i in use_spectra_B):
                rvA = radial_velocity_collection_A[i]
                rvB = radial_velocity_collection_B[i]
                if options.rv_lower_limit == 0.0 and use_spectra_B is None:
                    condition = np.abs(rvA-rvB) > options.rv_proximity_limit
                elif use_spectra_B is None:
                    condition = np.abs(rvA) > options.rv_lower_limit
                else:
                    condition = True

                if condition:
                    shifted_flux_B = shift_spectrum(flux_collection[:, i], -rvB, delta_v)
                    separated_flux_B += weights[i] * \
                                        (shifted_flux_B - shift_spectrum(separated_flux_A, rvA - rvB, delta_v))
                    n_used_spectra += weights[i]
            elif use_spectra_B.size != 0:
                pass
            else:
                raise TypeError(f'use_spectra_B is either of wrong type ({type(use_spectra_B)}), empty, or wrong value.\n' +
                                f'Expected type: {type(True)} or np.ndarray. Expected value if bool: True')
        separated_flux_B = separated_flux_B / n_used_spectra

        RMS_values_A += separated_flux_A
        RMS_values_B += separated_flux_B
        RMS_A = np.sum(RMS_values_A**2)/RMS_values_A.size
        RMS_B = np.sum(RMS_values_B**2)/RMS_values_B.size
        if RMS_A < options.convergence_limit and RMS_B < options.convergence_limit:
            if options.verbose is True:
                print(f'Separate Component Spectra: Convergence limit of {options.convergence_limit} successfully '
                      f'reached in {iteration_counter} iterations. \nReturning last separated spectra.')
            break
        elif iteration_counter >= options.max_iterations:
            warnings.warn(f'Warning: Iteration limit of {options.max_iterations} reached without reaching convergence '
                          f'limit of {options.convergence_limit}. \nCurrent RMS_A: {RMS_A}. RMS_B: {RMS_B} \n'
                          'Returning last separated spectra.')
            break
    if options.verbose is True:
        print('n_spectra vs n_used_spectra: ', n_spectra, ' ', n_used_spectra)
    separated_flux_B += 1.0     # add back continuum to second component
    return separated_flux_A, separated_flux_B


def separate_component_spectra_orders(
        flux_order_collection, radial_velocity_collection_A, radial_velocity_collection_B,
        options: SeparateComponentsOptions, overlap_weights
):
    """
    Assumes flux_order_collection.shape = [wavelength.size, n_orders, n_spectra], where wavelength.size is the amount of
    elements in the equi-velocity-spaced interpolated wavelength grid for a merged spectrum.
    Each individual order should have the same amount of elements, but values outside the range must be = 1
    mask_collection_orders.shape = [wavelength.size, n_orders, n_spectra]. dtype=bool. Each mask must remove values
    outside of order range
    """

    n_spectra = flux_order_collection[0, 0, :].size
    n_orders = flux_order_collection[0, :, 0].size
    separated_flux_A = np.zeros((flux_order_collection[:, 0, 0].size, ))
    separated_flux_B = np.zeros((flux_order_collection[:, 0, 0].size, ))
    use_spectra_A = options.use_for_spectral_separation_A
    use_spectra_B = options.use_for_spectral_separation_B
    delta_v = options.delta_v
    weights = options.weights
    if weights is None:
        weights = np.ones(n_spectra)

    iteration_counter = 0
    while True:
        if iteration_counter == 0:
            RMS_values_A = np.zeros(shape=separated_flux_A.shape); RMS_values_B = np.zeros(shape=separated_flux_B.shape)
        else:
            RMS_values_A = -separated_flux_A
            RMS_values_B = -separated_flux_B
        iteration_counter += 1
        separated_flux_A[:] = 0.0
        n_used_spectra = 0
        for i in range(0, n_spectra):
            if (use_spectra_A is None) or (i in use_spectra_A):
                if radial_velocity_collection_A.ndim == 2:
                    rvA = np.mean(radial_velocity_collection_A[:, i])
                    rvB = np.mean(radial_velocity_collection_B[:, i])
                else:
                    rvA = radial_velocity_collection_A[i]
                    rvB = radial_velocity_collection_B[i]
                if options.rv_lower_limit == 0.0 and use_spectra_A is None:
                    condition = np.abs(rvA - rvB) > options.rv_proximity_limit
                elif use_spectra_A is None:
                    condition = np.abs(rvA) > options.rv_lower_limit
                else:
                    condition = True
                if condition:
                    overlap_mean = np.zeros((separated_flux_A.size, ))
                    for j in range(0, n_orders):
                        if radial_velocity_collection_A.ndim == 2:
                            rvA = radial_velocity_collection_A[j, i]
                            rvB = radial_velocity_collection_B[j, i]
                        shifted_flux_A = shift_spectrum((flux_order_collection[:, j, i])*overlap_weights[:, j, i],
                                                        -rvA, delta_v)
                        if options.ignore_component_B is False:
                            shift_sep_B = shift_spectrum(overlap_weights[:, j, i]*separated_flux_B, rvB-rvA, delta_v)
                            separated_flux_A += (weights[i]*(shifted_flux_A - shift_sep_B))
                        else:
                            separated_flux_A += weights[i] * shifted_flux_A
                        overlap_mean += shift_spectrum(overlap_weights[:, j, i], -rvA, delta_v)
                    n_used_spectra += weights[i] * overlap_mean
            elif use_spectra_A.size != 0:
                pass
            else:
                raise TypeError(
                    f'use_spectra_A is either of wrong type ({type(use_spectra_A)}), empty, or wrong value.\n' +
                    f'Expected type: {type(True)} or np.ndarray. Expected value if bool: True'
                )
        separated_flux_A = separated_flux_A / n_used_spectra

        separated_flux_B[:] = 0.0
        n_used_spectra = 0
        for i in range(0, n_spectra):
            if (use_spectra_B is None) or (i in use_spectra_B):
                if radial_velocity_collection_A.ndim == 2:
                    rvA = np.mean(radial_velocity_collection_A[:, i])
                    rvB = np.mean(radial_velocity_collection_B[:, i])
                else:
                    rvA = radial_velocity_collection_A[i]
                    rvB = radial_velocity_collection_B[i]
                if options.rv_lower_limit == 0.0 and use_spectra_B is None:
                    condition = np.abs(rvA-rvB) > options.rv_proximity_limit
                elif use_spectra_B is None:
                    condition = np.abs(rvA) > options.rv_proximity_limit
                else:
                    condition = True

                if condition:
                    overlap_mean = np.zeros((separated_flux_B.size,))
                    for j in range(0, n_orders):
                        if radial_velocity_collection_A.ndim == 2:
                            rvA = radial_velocity_collection_A[j, i]
                            rvB = radial_velocity_collection_B[j, i]
                        shifted_flux_B = shift_spectrum((flux_order_collection[:, j, i])*overlap_weights[:, j, i],
                                                        -rvB, delta_v)
                        shift_sep_A = shift_spectrum(overlap_weights[:, j, i]*separated_flux_A, rvB-rvA, delta_v)
                        separated_flux_B += weights[i] * (shifted_flux_B - shift_sep_A)
                        overlap_mean += shift_spectrum(overlap_weights[:, j, i], -rvB, delta_v)
                    n_used_spectra += weights[i] * overlap_mean
            elif use_spectra_B.size != 0:
                pass
            else:
                raise TypeError(
                    f'use_spectra_B is either of wrong type ({type(use_spectra_B)}), empty, or wrong value.\n' +
                    f'Expected type: {type(True)} or np.ndarray. Expected value if bool: True')

        separated_flux_B = separated_flux_B/n_used_spectra

        RMS_values_A += separated_flux_A
        RMS_values_B += separated_flux_B
        RMS_values_A = RMS_values_A[~np.isnan(RMS_values_A)]
        RMS_values_B = RMS_values_B[~np.isnan(RMS_values_B)]
        RMS_A = np.sum(RMS_values_A ** 2) / RMS_values_A.size
        RMS_B = np.sum(RMS_values_B ** 2) / RMS_values_B.size
        if RMS_A < options.convergence_limit and RMS_B < options.convergence_limit:
            if options.verbose is True:
                print(f'Separate Component Spectra: Convergence limit of {options.convergence_limit} successfully '
                      f'reached in {iteration_counter} iterations. \nReturning last separated spectra.')
            break
        elif iteration_counter >= options.max_iterations:
            warnings.warn(
                f'Warning: Iteration limit of {options.max_iterations} reached without reaching convergence '
                f'limit of {options.convergence_limit}. \nCurrent RMS_A: {RMS_A}. RMS_B: {RMS_B} \n'
                'Returning last separated spectra.')
            break
    separated_flux_B += 1.0     # add back a continuum to second component
    separated_flux_A[np.isnan(separated_flux_A)] = 1.0
    separated_flux_B[np.isnan(separated_flux_B)] = 1.0
    return separated_flux_A, separated_flux_B


def _calculate_overlap(mask_collection_orders):
    overlap_weights = np.zeros(shape=mask_collection_orders.shape, dtype=float)
    n_orders = mask_collection_orders[0, :, 0].size
    n_spectra = mask_collection_orders[0, 0, :].size
    for i in range(0, n_spectra):
        for j in range(0, n_orders):
            other_masks = np.delete(mask_collection_orders[:, :, i], j, axis=1)
            mask = mask_collection_orders[:, j, i]
            overlap_mask = np.zeros(mask.size, dtype=bool)
            for k in range(0, n_orders-1):
                other_mask_temp = other_masks[:, k]
                comb_mask = mask & other_mask_temp
                overlap_mask = overlap_mask + comb_mask
            overlap_weights[(mask & ~overlap_mask), j, i] = 1.0
            mask_idx = np.argwhere(mask)
            mask_lowhalf = np.copy(mask)
            mask_lowhalf[mask_idx[mask_idx.size//2:]] = False       # Mask identifies only first half of data as True
            mask_highhalf = np.copy(mask)
            mask_highhalf[mask_idx[0:mask_idx.size//2]] = False
            overlap_low = overlap_mask & mask_lowhalf
            overlap_high = overlap_mask & mask_highhalf
            overlap_weights[overlap_low, j, i] = np.linspace(0, 1, overlap_low[overlap_low].size)
            overlap_weights[overlap_high, j, i] = np.linspace(1, 0, overlap_high[overlap_high].size)
    return overlap_weights


def _update_bf_plot(plot_ax, model, RV_actual, get_fit_parameter_values, index):
    fit = model[0]
    model_values = model[1]
    velocity_values = model[2]
    bf_smooth_values = model[4]
    _, RV_measured, _, _, _, _ = get_fit_parameter_values(fit.params)
    RV_offset = RV_actual - RV_measured
    plot_ax.plot(velocity_values + RV_offset, 1+0.02*bf_smooth_values/np.max(bf_smooth_values)-0.05*index)
    plot_ax.plot(velocity_values + RV_offset, 1+0.02*model_values/np.max(bf_smooth_values)-0.05*index, 'k--')
    plot_ax.plot(np.ones(shape=(2,))*RV_actual,
                 [1-0.05*index-0.005, 1+0.025*np.max(model_values)/np.max(bf_smooth_values)-0.05*index],
                 color='grey')


def recalculate_RVs(
        flux_collection: np.ndarray, separated_flux_A: np.ndarray, separated_flux_B: np.ndarray,
        RV_collection_A: np.ndarray, RV_collection_B: np.ndarray, flux_templateA: np.ndarray,
        flux_templateB: np.ndarray, buffer_mask: np.ndarray, options: RadialVelocityOptions, get_fit_parameter_values,
        plot_ax_A=None, plot_ax_B=None, time_values=None, period=None
):
    """
    This part of the spectral separation routine corrects the spectra for the separated spectra found by
    separate_component_spectra and recalculates RV values for each component using the corrected spectra (with one
    component removed).

    :param flux_collection: np.ndarray shape (:, n_spectra). Collection of normalized fluxes for the program spectra.
    :param separated_flux_A:    np.ndarray shape (:, ). Meaned normalized flux from separate_component_spectra() for A.
    :param separated_flux_B:    np.ndarray shape (:, ). Meaned normalized flux from separate_component_spectra() for B.
    :param RV_collection_A:     np.ndarray shape (n_spectra, ). Current RV values used to remove A from spectrum with.
    :param RV_collection_B:     np.ndarray shape (n_spectra, ). Current RV values used to remove B from spectrum with.
    :param flux_templateA:  np.ndarray shape (:, ). Template spectrum normalized flux for component A.
    :param flux_templateB:  np.ndarray shape (:, ). Template spectrum normalized flux for component B.
    :param buffer_mask:         np.ndarray shape (:, ). Mask used to remove "buffer" (or "padding") from spectrum.
                                    See spectral_separation_routine() for more info.
    :param options:             fitting and broadening function options
    :param plot_ax_A:           matplotlib.axes.axes. Used to update RV plots during iterations.
    :param plot_ax_B:           matplotlib.axes.axes. Used to update RV plots during iterations.

    :return:    RV_collection_A,        RV_collection_B, (fits_A, fits_B)
                RV_collection_A:        updated values for the RV of component A.
                RV_collection_B:        updated values for the RV of component B.
                fits_A, fits_B:         np.ndarrays storing the rotational BF profile fits found for each spectrum.
    """
    RV_collection_A = deepcopy(RV_collection_A)
    RV_collection_B = deepcopy(RV_collection_B)
    n_spectra = flux_collection[0, :].size
    v_span = options.bf_velocity_span
    delta_v = options.delta_v

    if plot_ax_A is not None:
        plot_ax_A.clear()
        plot_ax_A.set_xlim([-v_span/2, +v_span/2])
        plot_ax_A.set_xlabel('Velocity shift [km/s]')
    if plot_ax_B is not None:
        plot_ax_B.clear()
        plot_ax_B.set_xlim([-v_span/2, +v_span/2])
        plot_ax_B.set_xlabel('Velocity shift [km/s]')

    bf_fitres_A = np.empty(shape=(n_spectra,), dtype=tuple)
    bf_fitres_B = np.empty(shape=(n_spectra,), dtype=tuple)

    krange = 1

    for i in range(0, n_spectra):
        vary_continuum = True
        continuum_A = 0.0
        continuum_B = 0.0
        if options.refit_width_A is not None or options.refit_width_B is not None:
            krange = 2
        else:
            pass
        for k in range(0, krange):       # perform "burn-in" with wide fit-width, and then refit only with peak data
            if k == 0:
                fit_width_A = copy(options.velocity_fit_width_A)
                fit_width_B = copy(options.velocity_fit_width_B)
            if k == 1:
                options.velocity_fit_width_A = options.refit_width_A
                options.velocity_fit_width_B = options.refit_width_B
                vary_continuum = False
                _, _, _, _, continuum_A, _ = get_fit_parameter_values(model_A[0].params)
                _, _, _, _, continuum_B, _ = get_fit_parameter_values(model_B[0].params)

            iterations = 0
            while True:
                iterations += 1

                # # Calculate RV_A # #
                corrected_flux_A = (1-flux_collection[:, i]) - \
                                   shift_spectrum(1-separated_flux_B, RV_collection_B[i], delta_v)

                if period is not None and options.ignore_at_phase_B is not None and time_values is not None:
                    if _check_for_total_eclipse(time_values[i], period, options.ignore_at_phase_B) is True:
                        corrected_flux_A = 1-flux_collection[:, i]

                corrected_flux_A = corrected_flux_A[~buffer_mask]

                # Generate fit parameter object
                fitparams_A = FitParameters(
                    options.vsini_A, options.spectral_resolution, options.velocity_fit_width_A, options.limbd_coef_A,
                    options.bf_smooth_sigma_A, options.bf_velocity_span, options.vary_vsini_A,
                    options.vsini_vary_limit_A, options.vary_limbd_coef_A, RV=0.0, continuum=continuum_A,
                    vary_continuum=vary_continuum
                )

                # Perform calculation
                template_shifted = shift_spectrum(flux_templateA, RV_collection_A[i], delta_v)
                BRsvd_template_A = BroadeningFunction(  # observation flux is changed to corrected_flux_A later
                    corrected_flux_A, 1 - template_shifted[~buffer_mask], v_span, delta_v
                )
                BRsvd_template_A.smooth_sigma = options.bf_smooth_sigma_A
                RV_deviation_A, model_A = radial_velocity_single_component(
                    corrected_flux_A, BRsvd_template_A, fitparams_A
                )
                RV_collection_A[i] = RV_collection_A[i] + RV_deviation_A

                # # Calculate RV_B # #
                corrected_flux_B = (1-flux_collection[:, i]) - \
                                   shift_spectrum(1-separated_flux_A, RV_collection_A[i], delta_v)

                if period is not None and options.ignore_at_phase_A is not None and time_values is not None:
                    if _check_for_total_eclipse(time_values[i], period, options.ignore_at_phase_A) is True:
                        corrected_flux_B = 1-flux_collection[:, i]

                corrected_flux_B = corrected_flux_B[~buffer_mask]

                fitparams_B = FitParameters(
                    options.vsini_B, options.spectral_resolution, options.velocity_fit_width_B, options.limbd_coef_B,
                    options.bf_smooth_sigma_B, options.bf_velocity_span, options.vary_vsini_B,
                    options.vsini_vary_limit_B, options.vary_limbd_coef_B, RV=0.0, continuum=continuum_B,
                    vary_continuum=vary_continuum
                )
                template_shifted = shift_spectrum(flux_templateB, RV_collection_B[i], delta_v)
                BRsvd_template_B = BroadeningFunction(
                    corrected_flux_B, 1 - template_shifted[~buffer_mask], v_span, delta_v
                )
                BRsvd_template_B.smooth_sigma = options.bf_smooth_sigma_B
                RV_deviation_B, model_B = radial_velocity_single_component(
                    corrected_flux_B, BRsvd_template_B, fitparams_B
                )
                RV_collection_B[i] = RV_collection_B[i] + RV_deviation_B
                if options.verbose:
                    print(
                        f'RV dev spec {i}: {np.abs(RV_deviation_A):.{options.print_prec}f} (A) '
                        f'{np.abs(RV_deviation_B):.{options.print_prec}f} (B)'
                    )
                if (np.abs(RV_deviation_A) < options.convergence_limit or i not in options.evaluate_spectra_A) and \
                        (np.abs(RV_deviation_B) < options.convergence_limit or i not in options.evaluate_spectra_B):
                    if options.verbose:
                        print(f'RV: spectrum {i} successful.')
                    break
                elif iterations > options.iteration_limit:
                    if k == 1 and options.verbose is True:
                        warnings.warn(
                            f'RV: spectrum {i} did not reach convergence limit {options.convergence_limit}.'
                        )
                    break

        bf_fitres_A[i], bf_fitres_B[i] = model_A, model_B

        rv_lower_limit = options.rv_lower_limit
        if plot_ax_A is not None:
            _update_bf_plot(plot_ax_A, model_A, RV_collection_A[i], get_fit_parameter_values, i)
            if rv_lower_limit != 0.0:
                plot_ax_A.plot([rv_lower_limit, rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
                plot_ax_A.plot([-rv_lower_limit, -rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
        if plot_ax_B is not None:
            _update_bf_plot(plot_ax_B, model_B, RV_collection_B[i], get_fit_parameter_values, i)
            if rv_lower_limit != 0.0:
                plot_ax_B.plot([rv_lower_limit, rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
                plot_ax_B.plot([-rv_lower_limit, -rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)

    if krange == 2:
        options.velocity_fit_width_A = fit_width_A
        options.velocity_fit_width_B = fit_width_B

    return RV_collection_A, RV_collection_B, (bf_fitres_A, bf_fitres_B)


def recalculate_RVs_orders(
        flux_collection_orders, mask_collection_orders, separated_flux_A, separated_flux_B, RV_collection_orders_A,
        RV_collection_orders_B, flux_templateA, flux_templateB, options: RadialVelocityOptions,
        get_fit_parameter_values, plot_ax_A=None,
        plot_ax_B=None, time_values=None, period=None, plot_order=0
):
    """
    Assumes flux_order_collection.shape = [wavelength.size, n_orders, n_spectra], where wavelength.size is the amount of
    elements in the equi-velocity-spaced interpolated wavelength grid for a merged spectrum.
    Each individual order should have the same amount of elements, but values outside the range must be = 1
    mask_collection_orders.shape = [wavelength.size, n_orders, n_spectra]. dtype=bool. Each mask must remove values
    outside of order range
    RV_collection_orders_X.shape = (n_orders, n_spectra)
    """
    RV_collection_orders_A = deepcopy(RV_collection_orders_A)
    RV_collection_orders_B = deepcopy(RV_collection_orders_B)
    n_spectra = flux_collection_orders[0, 0, :].size
    n_orders = flux_collection_orders[0, :, 0].size
    v_span = options.bf_velocity_span
    delta_v = options.delta_v
    krange = 1

    bf_fitres_A = np.empty(shape=(n_orders, n_spectra), dtype=tuple)
    bf_fitres_B = np.empty(shape=(n_orders, n_spectra), dtype=tuple)

    if options.n_parallel_jobs > 1:
        arguments = (
            RV_collection_orders_A, RV_collection_orders_B, flux_collection_orders, separated_flux_A,
            separated_flux_B, flux_templateA, flux_templateB, mask_collection_orders, delta_v, v_span, period,
            time_values, n_orders, krange, bf_fitres_A, bf_fitres_B, options
        )
        res_par = Parallel(n_jobs=options.n_parallel_jobs)(
            delayed(_rv_loop_orders)(*arguments, i) for i in range(0, n_spectra)
        )
        for i in range(0, n_spectra):
            result_temp = res_par[i]
            RV_collection_orders_A[:, i] = result_temp[0]
            RV_collection_orders_B[:, i] = result_temp[1]
            bf_fitres_A[:, i] = result_temp[2][0]
            bf_fitres_B[:, i] = result_temp[2][1]

            rv_lower_limit = options.rv_lower_limit
            if plot_ax_A is not None:
                _update_bf_plot(
                    plot_ax_A, bf_fitres_A[0, i], RV_collection_orders_A[0, i], get_fit_parameter_values, i
                )
                if rv_lower_limit != 0.0:
                    plot_ax_A.plot([rv_lower_limit, rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
                    plot_ax_A.plot([-rv_lower_limit, -rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
            if plot_ax_B is not None:
                _update_bf_plot(
                    plot_ax_B, bf_fitres_B[0, i], RV_collection_orders_B[0, i], get_fit_parameter_values, i
                )
                if rv_lower_limit != 0.0:
                    plot_ax_B.plot([rv_lower_limit, rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
                    plot_ax_B.plot([-rv_lower_limit, -rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)

    else:
        for i in range(0, n_spectra):
            if options.verbose:
                print(f'RV: spectrum {i} begun.')
            if options.refit_width_A is not None or options.refit_width_B is not None:
                krange = 2
            else:
                pass
            for k in range(0, krange):
                if k == 0:
                    fit_width_A = copy(options.velocity_fit_width_A)
                    fit_width_B = copy(options.velocity_fit_width_B)
                if k == 1:
                    options.velocity_fit_width_A = options.refit_width_A
                    options.velocity_fit_width_B = options.refit_width_B
                iterations = 0
                RV_deviation_A = np.ones((RV_collection_orders_A[:, 0].size,))
                RV_deviation_B = np.ones((RV_collection_orders_B[:, 0].size,))
                while True:
                    iterations += 1
                    for j in range(0, n_orders):
                        # # Calculate RV_A # #
                        corrected_flux_A = (1 - flux_collection_orders[:, j, i]) - \
                                           (1 - shift_spectrum(separated_flux_B, RV_collection_orders_B[j, i], delta_v))
                        if period is not None and options.ignore_at_phase_B is not None and time_values is not None:
                            if _check_for_total_eclipse(time_values[i], period, options.ignore_at_phase_B) is True:
                                corrected_flux_A = 1 - flux_collection_orders[:, j, i]
                        corrected_flux_A = corrected_flux_A[mask_collection_orders[:, j, i]]
                        fitparams_A = FitParameters(
                            options.vsini_A, options.spectral_resolution, options.velocity_fit_width_A,
                            options.limbd_coef_A,
                            options.bf_smooth_sigma_A, options.bf_velocity_span, options.vary_vsini_A,
                            options.vsini_vary_limit_A, options.vary_limbd_coef_A, RV=0.0
                        )
                        template_shifted = shift_spectrum(  # shift template to reduce boundary issues
                            flux_templateA, RV_collection_orders_A[j, i], delta_v
                        )[mask_collection_orders[:, j, i]]
                        BRsvd_template_A = BroadeningFunction(
                            corrected_flux_A,
                            1 - template_shifted,
                            v_span, delta_v
                        )
                        BRsvd_template_A.smooth_sigma = options.bf_smooth_sigma_A
                        rva_temp, bf_fitres_A[j, i] = radial_velocity_single_component(
                            corrected_flux_A, BRsvd_template_A, fitparams_A
                        )
                        RV_deviation_A[j] = rva_temp
                        RV_collection_orders_A[j, i] = rva_temp + RV_collection_orders_A[j, i]

                        # # Calculate RV_B # #
                        corrected_flux_B = (1 - flux_collection_orders[:, j, i]) - \
                                           (1 - shift_spectrum(separated_flux_A, RV_collection_orders_A[j, i], delta_v))
                        if period is not None and options.ignore_at_phase_A is not None and time_values is not None:
                            if _check_for_total_eclipse(time_values[i], period, options.ignore_at_phase_A) is True:
                                corrected_flux_B = 1 - flux_collection_orders[:, j, i]
                        corrected_flux_B = corrected_flux_B[mask_collection_orders[:, j, i]]
                        fitparams_B = FitParameters(
                            options.vsini_B, options.spectral_resolution, options.velocity_fit_width_B,
                            options.limbd_coef_B,
                            options.bf_smooth_sigma_B, options.bf_velocity_span, options.vary_vsini_B,
                            options.vsini_vary_limit_B, options.vary_limbd_coef_B, RV=0.0
                        )
                        template_shifted = shift_spectrum(
                            flux_templateB[mask_collection_orders[:, j, i]], RV_collection_orders_B[j, i], delta_v
                        )
                        BRsvd_template_B = BroadeningFunction(
                            corrected_flux_B,
                            1 - template_shifted, v_span, delta_v
                        )
                        BRsvd_template_B.smooth_sigma = options.bf_smooth_sigma_B
                        rvb_temp, bf_fitres_B[j, i] = radial_velocity_single_component(
                            corrected_flux_B, BRsvd_template_B, fitparams_B
                        )
                        RV_deviation_B[j] = rvb_temp
                        RV_collection_orders_B[j, i] = rvb_temp + RV_collection_orders_B[j, i]
                    if options.verbose:
                        print(
                            f'RV dev spec {i}: {np.abs(np.mean(RV_deviation_A)):.{options.print_prec}f} (A) '
                            f'{np.abs(np.mean(RV_deviation_B)):.{options.print_prec}f} (B)'
                        )
                    if (np.abs(np.mean(RV_deviation_A)) < options.convergence_limit or i not in
                        options.evaluate_spectrum_A) and (np.abs(np.mean(RV_deviation_B)) < options.convergence_limit
                                                          or i not in options.evaluate_spectrum_B):
                        if options.verbose:
                            print(f'RV: spectrum {i} successful.')
                        break
                    elif iterations > options.iteration_limit:
                        if k == 1 or (options.refit_width_B is None and options.refit_width_A is None):
                            if options.verbose is True:
                                warnings.warn(
                                    f'RV: spectrum {i} did not reach convergence limit {options.convergence_limit}.'
                                )
                        break

            rv_lower_limit = options.rv_lower_limit
            if plot_ax_A is not None:
                _update_bf_plot(plot_ax_A, bf_fitres_A[plot_order, i], RV_collection_orders_A[plot_order, i], i)
                if rv_lower_limit != 0.0:
                    plot_ax_A.plot([rv_lower_limit, rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
                    plot_ax_A.plot([-rv_lower_limit, -rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
            if plot_ax_B is not None:
                _update_bf_plot(plot_ax_B, bf_fitres_B[plot_order, i], RV_collection_orders_B[plot_order, i], i)
                if rv_lower_limit != 0.0:
                    plot_ax_B.plot([rv_lower_limit, rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)
                    plot_ax_B.plot([-rv_lower_limit, -rv_lower_limit], [0, 1.1], 'k', linewidth=0.3)

    if krange == 2 and options.n_parallel_jobs == 1:
        options.velocity_fit_width_A = fit_width_A
        options.velocity_fit_width_B = fit_width_B

    return RV_collection_orders_A, RV_collection_orders_B, (bf_fitres_A, bf_fitres_B)


def _rv_loop_orders(
        RV_collection_orders_A, RV_collection_orders_B, flux_collection_orders, separated_flux_A, separated_flux_B,
        flux_templateA, flux_templateB, mask_collection_orders, delta_v, v_span, period, time_values, n_orders, krange,
        bf_fitres_A, bf_fitres_B, options,
        i
):
    # RV_collection_orders_A = np.copy(RV_collection_orders_A)
    # RV_collection_orders_B = np.copy(RV_collection_orders_B)
    # bf_fitres_A = np.copy(bf_fitres_A)
    # bf_fitres_B = np.copy(bf_fitres_B)
    options = deepcopy(options) # probably not needed
    if options.verbose:
        print(f'RV: spectrum {i} begun.')
    if options.refit_width_A is not None or options.refit_width_B is not None:
        krange = 2
    else:
        pass
    for k in range(0, krange):
        if k == 0:
            fit_width_A = copy(options.velocity_fit_width_A)
            fit_width_B = copy(options.velocity_fit_width_B)
        if k == 1:
            options.velocity_fit_width_A = options.refit_width_A
            options.velocity_fit_width_B = options.refit_width_B
        iterations = 0
        RV_deviation_A = np.ones((RV_collection_orders_A[:, 0].size,))
        RV_deviation_B = np.ones((RV_collection_orders_B[:, 0].size,))
        while True:
            iterations += 1
            for j in range(0, n_orders):
                # # Calculate RV_A # #
                corrected_flux_A = (1 - flux_collection_orders[:, j, i]) - \
                                   (1 - shift_spectrum(separated_flux_B, RV_collection_orders_B[j, i], delta_v))
                if period is not None and options.ignore_at_phase_B is not None and time_values is not None:
                    if _check_for_total_eclipse(time_values[i], period, options.ignore_at_phase_B) is True:
                        corrected_flux_A = 1 - flux_collection_orders[:, j, i]
                corrected_flux_A = corrected_flux_A[mask_collection_orders[:, j, i]]

                fitparams_A = FitParameters(
                    options.vsini_A, options.spectral_resolution, options.velocity_fit_width_A, options.limbd_coef_A,
                    options.bf_smooth_sigma_A, options.bf_velocity_span, options.vary_vsini_A,
                    options.vsini_vary_limit_A, options.vary_limbd_coef_A, RV=0.0
                )
                template_shifted = shift_spectrum(  # shift template to reduce boundary issues
                    flux_templateA, RV_collection_orders_A[j, i], delta_v
                )[mask_collection_orders[:, j, i]]
                BRsvd_template_A = BroadeningFunction(
                    corrected_flux_A,
                    1 - template_shifted,
                    v_span, delta_v
                )
                BRsvd_template_A.smooth_sigma = options.bf_smooth_sigma_A
                rva_temp, bf_fitres_A[j, i] = radial_velocity_single_component(
                    corrected_flux_A, BRsvd_template_A, fitparams_A
                )
                RV_deviation_A[j] = rva_temp
                RV_collection_orders_A[j, i] = rva_temp + RV_collection_orders_A[j, i]

                # # Calculate RV_B # #
                corrected_flux_B = (1 - flux_collection_orders[:, j, i]) - \
                                   (1 - shift_spectrum(separated_flux_A, RV_collection_orders_A[j, i], delta_v))
                if period is not None and options.ignore_at_phase_A is not None and time_values is not None:
                    if _check_for_total_eclipse(time_values[i], period, options.ignore_at_phase_A) is True:
                        corrected_flux_B = 1 - flux_collection_orders[:, j, i]
                corrected_flux_B = corrected_flux_B[mask_collection_orders[:, j, i]]
                options.RV_B = RV_collection_orders_B[j, i]
                fitparams_B = FitParameters(
                    options.vsini_B, options.spectral_resolution, options.velocity_fit_width_B, options.limbd_coef_B,
                    options.bf_smooth_sigma_B, options.bf_velocity_span, options.vary_vsini_B,
                    options.vsini_vary_limit_B, options.vary_limbd_coef_B, RV=0.0
                )
                template_shifted = shift_spectrum(
                    flux_templateB[mask_collection_orders[:, j, i]], RV_collection_orders_B[j, i], delta_v
                )
                BRsvd_template_B = BroadeningFunction(
                    corrected_flux_B,
                    1 - template_shifted, v_span, delta_v
                )
                BRsvd_template_B.smooth_sigma = options.bf_smooth_sigma_B
                rvb_temp, bf_fitres_B[j, i] = radial_velocity_single_component(
                    corrected_flux_B, BRsvd_template_B, fitparams_B
                )
                RV_deviation_B[j] = rvb_temp
                RV_collection_orders_B[j, i] = rvb_temp + RV_collection_orders_B[j, i]
            if options.verbose:
                print(
                    f'RV dev spec {i}: {np.abs(np.mean(RV_deviation_A)):.{options.print_prec}f} (A) '
                    f'{np.abs(np.mean(RV_deviation_B)):.{options.print_prec}f} (B)'
                )
            if (np.abs(np.mean(RV_deviation_A)) < options.convergence_limit or i not in options.evaluate_spectrum_A) \
                    and (np.abs(np.mean(RV_deviation_B)) < options.convergence_limit or i not in
                         options.evaluate_spectrum_B):
                if options.verbose:
                    print(f'RV: spectrum {i} successful.')
                break
            elif iterations > options.iteration_limit:
                if k == 1 or (options.refit_width_B is None and options.refit_width_A is None):
                    if options.verbose is True:
                        warnings.warn(
                            f'RV: spectrum {i} did not reach convergence limit {options.convergence_limit}.'
                        )
                break
    return RV_collection_orders_A[:, i], RV_collection_orders_B[:, i], (bf_fitres_A[:, i], bf_fitres_B[:, i])


def _check_for_total_eclipse(time_value, period, eclipse_phase_area):
    phase = np.mod(time_value, period)/period
    lower = eclipse_phase_area[0]
    upper = eclipse_phase_area[1]
    if lower < upper:
        condition = (phase > lower) & (phase < upper)
    elif lower > upper:
        condition = (phase > lower) | (phase < upper)
    else:
        raise ValueError('eclipse_phase_area must comprise of a lower and an upper value that are separate.')
    return condition


def _initialize_ssr_plots():
    # RVs and separated spectra
    fig_1 = plt.figure(figsize=(16, 9))
    gs_1 = fig_1.add_gridspec(2, 2)
    f1_ax1 = fig_1.add_subplot(gs_1[0, :])
    f1_ax2 = fig_1.add_subplot(gs_1[1, 0])
    f1_ax3 = fig_1.add_subplot(gs_1[1, 1])

    # Broadening function fits A
    fig_3 = plt.figure(figsize=(16, 9))
    gs_3 = fig_3.add_gridspec(1, 1)
    f3_ax1 = fig_3.add_subplot(gs_3[:, :])

    # Broadening function fits B
    fig_4 = plt.figure(figsize=(16, 9))
    gs_4 = fig_4.add_gridspec(1, 1)
    f4_ax1 = fig_4.add_subplot(gs_4[:, :])
    return f1_ax1, f1_ax2, f1_ax3, f3_ax1, f4_ax1


def _plot_ssr_iteration(
        f1_ax1, f1_ax2, f1_ax3, separated_flux_A, separated_flux_B, wavelength, flux_template_A,
        flux_template_B, RV_collection_A, RV_collection_B, time, period, buffer_mask, rv_lower_limit,
        rv_proximity_limit
):
    f1_ax1.clear(); f1_ax2.clear(); f1_ax3.clear()
    separated_flux_A, separated_flux_B = separated_flux_A[~buffer_mask], separated_flux_B[~buffer_mask]
    wavelength = wavelength[~buffer_mask]
    flux_template_A, flux_template_B = flux_template_A[~buffer_mask], flux_template_B[~buffer_mask]

    if RV_collection_A.ndim == 2:
        RV_A = np.mean(RV_collection_A, axis=0)
        RV_B = np.mean(RV_collection_B, axis=0)
    else:
        RV_A = RV_collection_A
        RV_B = RV_collection_B
    if period is None:
        xval = time
        f1_ax1.set_xlabel('BJD - 245000')
    else:
        xval = np.mod(time, period) / period
        f1_ax1.set_xlabel('Phase')
    if rv_lower_limit == 0.0:
        RV_below_limit_mask = np.abs(RV_A - RV_B) < rv_proximity_limit
    else:
        RV_below_limit_mask = np.abs(RV_A) < rv_lower_limit
    f1_ax1.plot(xval[~RV_below_limit_mask], RV_A[~RV_below_limit_mask], 'b*')
    f1_ax1.plot(xval[~RV_below_limit_mask], RV_B[~RV_below_limit_mask], 'r*')
    f1_ax1.plot(xval[RV_below_limit_mask], RV_A[RV_below_limit_mask], 'bx')
    f1_ax1.plot(xval[RV_below_limit_mask], RV_B[RV_below_limit_mask], 'rx')

    f1_ax2.plot(wavelength, flux_template_A, '--', color='grey', linewidth=0.25)
    f1_ax2.plot(wavelength, separated_flux_A, 'b', linewidth=2)

    f1_ax3.plot(wavelength, flux_template_B, '--', color='grey', linewidth=0.25)
    f1_ax3.plot(wavelength, separated_flux_B, 'r', linewidth=2)

    f1_ax1.set_ylabel('Radial Velocity [km/s]')
    f1_ax2.set_ylabel('Normalized Flux')
    f1_ax2.set_xlabel('Wavelength [Å]')
    f1_ax3.set_xlabel('Wavelength [Å]')
    f1_ax3.set_ylim([0.0, 1.15])
    f1_ax2.set_ylim([0.0, 1.15])

    plt.draw_all()
    plt.pause(1.5)


def save_multi_image(filename):
    """
    https://www.tutorialspoint.com/saving-all-the-open-matplotlib-figures-in-one-file-at-once
    """
    pp = PdfPages(filename)
    fig_nums = plt.get_fignums()
    figs = [plt.figure(n) for n in fig_nums]
    for fig in figs:
        fig.savefig(pp, format='pdf')
    pp.close()


def spectral_separation_routine(
        flux_collection: np.ndarray, flux_templateA: np.ndarray, flux_templateB: np.ndarray,
        wavelength: np.ndarray,
        options: RoutineOptions, sep_comp_options: SeparateComponentsOptions, rv_options: RadialVelocityOptions,
        RV_guess_collection: np.ndarray, mask_collection_orders=None, time_values: np.ndarray = None,
        period: float = None, buffer_mask: np.ndarray = None, outside_separated_A=None, outside_separated_B=None,
        weights_outside_A=None, weights_outside_B=None
):
    """
    Routine that separates component spectra and calculates radial velocities by iteratively calling
    separate_component_spectra() and recalculate_RVs() to attempt to converge towards the correct RV values for the
    system. Requires good starting guesses on the component RVs. It also plots each iteration (if enabled) to follow
    along the results.

    The routine must be provided with normalized spectra sampled in the same wavelengths equi-spaced in velocity space.
    It is recommended to supply spectra that are buffered (padded) in the ends, and a buffer_mask that indicates this.
    This will limit the effect of loop-back when rolling (shifting) the spectra. spf.limit_wavelength_interval() can be
    used to produce a buffered data-set (and others of the same size).

    The spectra used for calculating the separated spectra should cover a wide range of orbital phases to provide a
    decent averaged spectrum. They should have a high S/N, and should not include significant emission lines. Avoid
    using spectra with RVs of the two components close to each others, or spectra from within eclipses.

    Additional features included to improve either the RV fitting process or the spectral separation:
      - Fitted v*sin(i) values are meaned and used as fit guesses for next iteration. Simultaneously, the parameter
        is bounded to the limits v*sin(i) * [1-0.5, 1+0.5], in order to ensure stable fitting for all spectra.

      - The routine can be provided with an array of indices specifying which spectra to use when creating the separated
        spectra. This is the recommended way to designate. A lower RV limit on the primary component, or a proximity
        limit of the two RVs, can also be provided isntead. However, this does not consider any other important
        factors for why you would leave out a spectrum (low S/N, bad reduction, within eclipses).
        Providing rv_lower_limit (or rv_proximity_limit) while also supplying use_spectra, can be useful since the
        routines avoids using the "bad" component B RVs when calculating RMS values.
        Note that if use_spectra is provided as an array, the two rv limit options will only be used for RMS
        calculation, and not for spectral separation.

    :param flux_collection:   np.ndarray shape (:, n_spectra). Collection of normalized fluxes for each spectrum.
                                        if ndims=3, assumes shape (:, n_orders, n_spectra) containing independent orders
                                        Then mask_collection_orders must be supplied as well
    :param flux_templateA:    np.ndarray shape (:, ). Normalized flux of template spectrum for component A.
    :param flux_templateB:    np.ndarray shape (:, ). Normalized flux of template spectrum for component B.
    :param wavelength:            np.ndarray shape (:, ). Wavelength values for the spectra (both program and template).
    :param options:               Options for this global routine.
    :param sep_comp_options:      Options to pass on to the subroutine separate_component_spectra.
    :param rv_options:            Options to pass on to the subroutine recalculate_RVs.
    :param RV_guess_collection:   np.ndarray shape (n_spectra, 2). Initial RV guesses for each component (A: [:, 0]).
    :param buffer_mask: np.ndarray shape (:, ), Default = None. If data is padded with a buffer around the edges,
                        this mask should be supplied. It should give the unbuffered data when flipped, e.g.
                        `flux_buffered[~buffer_mask] = flux`.

    :return:    RV_collection_A,  RV_collection_B, separated_flux_A, separated_flux_B, wavelength
                RV_collection_A:  np.ndarray shape (n_spectra, ). RV values of component A for each program spectrum.
                RV_collection_B:  same, but for component B (includes values below rv_lower_limit).
                separated_flux_A: np.ndarray shape (:*, ). The found "separated" or "disentangled" spectrum for A.
                                    It is an inverted flux (1-normalized_flux).
                separated_flux_B: np.ndarray shape (:*, ). The found "separated" or "disentangled" spectrum for B.
                wavelength:       np.ndarray shape (:*, ). Wavelength values for the separated spectra.

            Note on :*
                if buffer_mask is provided, the returned spectra will be the un-buffered versions, meaning
                separated_flux_A.size = flux_templateA[buffer_mask].size. Same for the returned wavelength.
                This can be disabled by setting return_unbuffered=False in the options object.
    """
    if rv_options.fitting_routine == 'RotBF':
        from rotational_broadening_function_fitting import get_fit_parameter_values
    elif rv_options.fitting_routine == 'Gaussian':
        from gaussian_fitting import get_fit_parameter_values
    else:
        raise ValueError('Unrecognised fitting routine.')

    if RV_guess_collection.ndim == 2:
        RV_collection_A, RV_collection_B = np.copy(RV_guess_collection[:, 0]), np.copy(RV_guess_collection[:, 1])
    elif RV_guess_collection.ndim == 3:
        RV_collection_A = np.copy(RV_guess_collection[:, :, 0])
        RV_collection_B = np.copy(RV_guess_collection[:, :, 1])
    else:
        raise ValueError('RV_guess_collection.ndim != 2 or 3.')

    convergence_limit = options.convergence_limit
    iteration_limit = options.iteration_limit
    sep_comp_options.delta_v = options.delta_v
    rv_options.delta_v = options.delta_v

    if buffer_mask is None:
        buffer_mask = np.zeros(wavelength.shape, dtype=bool)

    # Initialize plot figures
    if options.plot:
        f1_ax1, f1_ax2, f1_ax3, f3_ax1, f4_ax1 = _initialize_ssr_plots()
    else:
        f1_ax1 = None; f1_ax2 = None; f1_ax3 = None; f3_ax1 = None; f4_ax1 = None

    if flux_collection.ndim == 3:
        overlap_weights = _calculate_overlap(mask_collection_orders)

    if rv_options.evaluate_spectra_A is None:
        if RV_collection_A.ndim == 1:
            rv_options.evaluate_spectra_A = np.linspace(0, RV_collection_A.size-1, RV_collection_A.size, dtype=int)
        else:
            rv_options.evaluate_spectra_A = np.linspace(
                0, RV_collection_A[0, :].size-1, RV_collection_A[0, :].size, dtype=int
            )
    if rv_options.evaluate_spectra_B is None:
        if RV_collection_B.ndim == 1:
            rv_options.evaluate_spectra_B = np.linspace(0, RV_collection_B.size-1, RV_collection_B.size, dtype=int)
        else:
            rv_options.evaluate_spectra_B = np.linspace(
                0, RV_collection_B[0, :].size-1, RV_collection_B[0, :].size, dtype=int
            )

    # Iterative loop that repeatedly separates the spectra from each other in order to calculate new RVs (Gonzales 2005)
    iterations = 0
    print('Spectral Separation: ')
    while True:
        print(f'\nIteration {iterations}.')
        if flux_collection.ndim == 3:
            if RV_collection_A.ndim == 1:
                RV_collection_A = RV_collection_A.reshape((1, RV_collection_A.size))
                RV_collection_B = RV_collection_B.reshape((1, RV_collection_B.size))
                RV_collection_A = np.repeat(RV_collection_A, axis=0, repeats=flux_collection[0, :, 0].size)
                RV_collection_B = np.repeat(RV_collection_B, axis=0, repeats=flux_collection[0, :, 0].size)
        if RV_collection_A.ndim == 2:
            rms_mask_A = np.zeros((RV_collection_A[0, :].size,), dtype=bool)
            rms_mask_B = np.zeros((RV_collection_B[0, :].size,), dtype=bool)
        else:
            rms_mask_A = np.zeros((RV_collection_A.size,), dtype=bool)
            rms_mask_B = np.zeros((RV_collection_B.size,), dtype=bool)

        if rv_options.evaluate_spectra_A is not None:
            rms_mask_A[rv_options.evaluate_spectra_A] = True
        else:
            rms_mask_A[:] = True

        if rv_options.evaluate_spectra_B is not None:
            rms_mask_B[rv_options.evaluate_spectra_B] = True
        elif sep_comp_options.rv_lower_limit == 0.0:
            if RV_collection_A.ndim == 2:
                rms_mask_B = np.abs(
                    np.mean(RV_collection_A, axis=0)-np.mean(RV_collection_B, axis=0)
                ) > sep_comp_options.rv_proximity_limit
            else:
                rms_mask_B = np.abs(RV_collection_A-RV_collection_B) > sep_comp_options.rv_proximity_limit
        else:
            if RV_collection_A.ndim == 2:
                rms_mask_B = np.abs(
                    np.mean(RV_collection_A, axis=0) - np.mean(RV_collection_B, axis=0)
                ) > sep_comp_options.rv_lower_limit
            else:
                rms_mask_B = np.abs(RV_collection_A) > sep_comp_options.rv_lower_limit

        if RV_collection_A.ndim == 2:
            RMS_A, RMS_B = -np.mean(RV_collection_A, axis=0)[rms_mask_A], -np.mean(RV_collection_B, axis=0)[rms_mask_B]
        else:
            RMS_A, RMS_B = -RV_collection_A[rms_mask_A], -RV_collection_B[rms_mask_B]

        if flux_collection.ndim == 3:
            separated_flux_A, separated_flux_B = separate_component_spectra_orders(
                flux_collection, RV_collection_A, RV_collection_B, sep_comp_options, overlap_weights
            )
        else:
            separated_flux_A, separated_flux_B = separate_component_spectra(
                flux_collection, RV_collection_A, RV_collection_B, sep_comp_options, outside_separated_A,
                outside_separated_B, weights_outside_A, weights_outside_B
            )

        if flux_collection.ndim == 3:
            RV_collection_A, RV_collection_B, (bf_fitres_A, bf_fitres_B) = recalculate_RVs_orders(
                flux_collection, mask_collection_orders, separated_flux_A, separated_flux_B, RV_collection_A,
                RV_collection_B, flux_templateA, flux_templateB, rv_options, get_fit_parameter_values,
                plot_ax_A=f3_ax1, plot_ax_B=f4_ax1,
                time_values=time_values, period=period, plot_order=options.plot_order
            )
        else:
            RV_collection_A, RV_collection_B, (bf_fitres_A, bf_fitres_B) = recalculate_RVs(
                flux_collection, separated_flux_A, separated_flux_B, RV_collection_A, RV_collection_B,
                flux_templateA, flux_templateB, buffer_mask, rv_options, get_fit_parameter_values, plot_ax_A=f3_ax1,
                plot_ax_B=f4_ax1, time_values=time_values, period=period
            )

        if options.plot:
            _plot_ssr_iteration(
                f1_ax1, f1_ax2, f1_ax3, separated_flux_A, separated_flux_B, wavelength, flux_templateA,
                flux_templateB, RV_collection_A, RV_collection_B, time_values, period,
                buffer_mask, sep_comp_options.rv_lower_limit, sep_comp_options.rv_proximity_limit
            )

        # Average vsini values for future fit guess and limit allowed fit area
        vsini_A, vsini_B = np.empty(shape=bf_fitres_A.shape), np.empty(shape=bf_fitres_B.shape)
        if vsini_A.ndim == 2:
            for i in range(0, vsini_A[0, :].size):
                for j in range(0, vsini_A[:, 0].size):
                    _, _, vsini_A[j, i], _, _, _ = get_fit_parameter_values(bf_fitres_A[j, i][0].params)
                    _, _, vsini_B[j, i], _, _, _ = get_fit_parameter_values(bf_fitres_B[j, i][0].params)
        else:
            for i in range(0, vsini_A.size):
                _, _, vsini_A[i], _, _, _ = get_fit_parameter_values(bf_fitres_A[i][0].params)
                _, _, vsini_B[i], _, _, _ = get_fit_parameter_values(bf_fitres_B[i][0].params)
        if options.adjust_vsini is True:
            rv_options.vsini_A = np.mean(vsini_A)
            rv_options.vsini_B = np.mean(vsini_B)
        if options.verbose:
            print('vsini A: ', rv_options.vsini_A)
            print('vsini B: ', rv_options.vsini_B)

        iterations += 1
        if RV_collection_A.ndim == 2:
            RMS_A += np.mean(RV_collection_A, axis=0)[rms_mask_A]
            RMS_B += np.mean(RV_collection_B, axis=0)[rms_mask_B]
            RMS_A = np.sqrt(np.sum(RMS_A**2)/np.mean(RV_collection_A, axis=0)[rms_mask_A].size)
            RMS_B = np.sqrt(np.sum(RMS_B**2)/np.mean(RV_collection_B, axis=0)[rms_mask_B].size)
        else:
            RMS_A += RV_collection_A[rms_mask_A]
            RMS_B += RV_collection_B[rms_mask_B]
            RMS_A = np.sqrt(np.sum(RMS_A**2))/RV_collection_A[rms_mask_A].size
            RMS_B = np.sqrt(np.sum(RMS_B**2))/RV_collection_B[rms_mask_B].size
        if options.verbose:
            print(f'RV_A RMS: {RMS_A}. ' + f'RV_B RMS: {RMS_B}.')
        if RMS_A < convergence_limit and RMS_B < convergence_limit:
            print(f'Spectral separation routine terminates after reaching convergence limit {convergence_limit}.')
            break
        if iterations >= iteration_limit:
            warnings.warn(f'RV convergence limit of {convergence_limit} not reached in {iterations} iterations.',
                          category=Warning)
            print('Spectral separation routine terminates.')
            break

    if options.save_all_results is True:
        save_separation_data(
            options.save_path, wavelength[~buffer_mask], time_values, RV_collection_A,
            RV_collection_B, RV_guess_collection, separated_flux_A[~buffer_mask], separated_flux_B[~buffer_mask],
            bf_fitres_A, bf_fitres_B, flux_templateA[~buffer_mask], flux_templateB[~buffer_mask], options
        )

    if options.save_plot_path is not None:
        save_multi_image(options.save_plot_path)

    if options.return_unbuffered:
        return RV_collection_A, RV_collection_B, separated_flux_A[~buffer_mask], separated_flux_B[~buffer_mask], \
               wavelength[~buffer_mask]
    else:
        return RV_collection_A, RV_collection_B, separated_flux_A, separated_flux_B, wavelength


def save_separation_data(
        location, wavelength, time_values, RVs_A, RVs_B, RVs_initial, separated_flux_A, separated_flux_B, bf_fitres_A,
        bf_fitres_B, template_flux_A, template_flux_B, options: RoutineOptions
):
    if options.filename_bulk is None:
        filename_bulk = str(int(np.round(np.min(wavelength)))) + '_' + str(int(np.round(np.max(wavelength))))
    else:
        filename_bulk = options.filename_bulk

    if RVs_A.ndim == 2:
        rvA_array = np.empty((RVs_A[0, :].size, 1+RVs_A[:, 0].size))
        rvA_array[:, 0] = time_values
        rvA_array[:, 1:] = RVs_A.T
    else:
        rvA_array = np.empty((RVs_A.size, 2))
        rvA_array[:, 0], rvA_array[:, 1] = time_values, RVs_A
    if RVs_B.ndim == 2:
        rvB_array = np.empty((RVs_B[0, :].size, 1+RVs_B[:, 0].size))
        rvB_array[:, 0] = time_values
        rvB_array[:, 1:] = RVs_B.T
    else:
        rvB_array = np.empty((RVs_B.size, 2))
        rvB_array[:, 0], rvB_array[:, 1] = time_values, RVs_B
    
    sep_array = np.empty((wavelength.size, 5))
    sep_array[:, 0], sep_array[:, 1], sep_array[:, 2] = wavelength, separated_flux_A, separated_flux_B
    sep_array[:, 3], sep_array[:, 4] = template_flux_A, template_flux_B

    if RVs_A.ndim == 2:
        fmt_tuple_A = ['%.6f' for x in range(0, RVs_A[:, 0].size)]
        fmt_tuple_A = ['%.9f'] + fmt_tuple_A
        fmt_tuple_A = tuple(fmt_tuple_A)
    else:
        fmt_tuple_A = ('%.9f', '%.6f')
    if RVs_B.ndim == 2:
        fmt_tuple_B = ['%.6f' for x in range(0, RVs_B[:, 0].size)]
        fmt_tuple_B = ['%.9f'] + fmt_tuple_B
        fmt_tuple_B = tuple(fmt_tuple_B)
    else:
        fmt_tuple_B = ('%.9f', '%.6f')
    np.savetxt(location + filename_bulk + '_rvA.txt', rvA_array, header='Time [input units] \t RV_A [km/s]',
               fmt=fmt_tuple_A)
    np.savetxt(location + filename_bulk + '_rv_initial.txt', RVs_initial, header='RV_A [km/s] \t RV_B [km/s]',
               fmt='%.6f')
    np.savetxt(location + filename_bulk + '_rvB.txt', rvB_array, header='Time [input units] \t RV_B [km/s]',
               fmt=fmt_tuple_B)
    np.savetxt(location + filename_bulk + '_sep_flux.txt', sep_array, header='Separated Flux A \t Separated Flux B \t Template A \t Template B', fmt='%.6f')

    if bf_fitres_A.ndim == 2:
        for j in range(0, bf_fitres_A[:, 0].size):
            vel_array = np.empty((bf_fitres_A[j, :].size, bf_fitres_A[j, 0][1].size))
            bf_array = np.empty((bf_fitres_A[j, :].size, bf_fitres_A[j, 0][1].size))
            bf_smooth_array = np.empty((bf_fitres_A[j, :].size, bf_fitres_A[j, 0][1].size))
            model_array = np.empty((bf_fitres_A[j, :].size, bf_fitres_A[j, 0][1].size))
            for i in range(0, bf_fitres_A[j, :].size):
                model_vals_A, bf_vel_A, bf_vals_A, smooth_vals_A = bf_fitres_A[j, i][1:]
                vel_array[i, :] = bf_vel_A
                bf_array[i, :] = bf_vals_A
                bf_smooth_array[i, :] = smooth_vals_A
                model_array[i, :] = model_vals_A
            np.savetxt(location+filename_bulk+'_order_'+str(j)+'_velocities_A.txt', vel_array, fmt='%.6f', header='BF velocities, shape (nspec, nvel)')
            np.savetxt(location + filename_bulk +'_order_'+str(j) + '_bfvals_A.txt', bf_array, fmt='%.6f',
                       header='BF vals, shape (nspec, nvel)')
            np.savetxt(location + filename_bulk +'_order_'+str(j) + '_bfsmooth_A.txt', bf_smooth_array, fmt='%.6f',
                       header='Smoothed BF vals, shape (nspec, nvel)')
            np.savetxt(location + filename_bulk +'_order_'+str(j) + '_models_A.txt', model_array, fmt='%.6f',
                       header='BF Fit vals, shape (nspec, nvel)')
    else:
        vel_array = np.empty((bf_fitres_A.size, bf_fitres_A[0][1].size))
        bf_array = np.empty((bf_fitres_A.size, bf_fitres_A[0][1].size))
        bf_smooth_array = np.empty((bf_fitres_A.size, bf_fitres_A[0][1].size))
        model_array = np.empty((bf_fitres_A.size, bf_fitres_A[0][1].size))
        for i in range(0, bf_fitres_A.size):
            model_vals_A, bf_velocity_A, bf_vals_A, bf_smooth_vals_A = bf_fitres_A[i][1:]
            vel_array[i, :] = bf_velocity_A
            bf_array[i, :] = bf_vals_A
            bf_smooth_array[i, :] = bf_smooth_vals_A
            model_array[i, :] = model_vals_A
        np.savetxt(location + filename_bulk + '_velocities_A.txt', vel_array, fmt='%.6f', header='BF velocities, shape (nspec, nvel)')
        np.savetxt(location + filename_bulk + '_bfvals_A.txt', bf_array, fmt='%.6f', header='BF vals, shape (nspec, nvel)')
        np.savetxt(location + filename_bulk + '_bfsmooth_A.txt', bf_smooth_array, fmt='%.6f',
               header='Smoothed BF vals, shape (nspec, nvel)')
        np.savetxt(location + filename_bulk + '_models_A.txt', model_array, fmt='%.6f', header='BF Fit vals, shape (nspec, nvel)')

    if bf_fitres_B.ndim == 2:
        for j in range(0, bf_fitres_B[:, 0].size):
            vel_array = np.empty((bf_fitres_B[j, :].size, bf_fitres_B[j, 0][1].size))
            bf_array = np.empty((bf_fitres_B[j, :].size, bf_fitres_B[j, 0][1].size))
            bf_smooth_array = np.empty((bf_fitres_B[j, :].size, bf_fitres_B[j, 0][1].size))
            model_array = np.empty((bf_fitres_B[j, :].size, bf_fitres_B[j, 0][1].size))
            for i in range(0, bf_fitres_B[j, :].size):
                model_vals_B, bf_vel_B, bf_vals_B, smooth_vals_B = bf_fitres_B[j, i][1:]
                vel_array[i, :] = bf_vel_B
                bf_array[i, :] = bf_vals_B
                bf_smooth_array[i, :] = smooth_vals_B
                model_array[i, :] = model_vals_B
            np.savetxt(location+filename_bulk+'_order_'+str(j)+'_velocities_B.txt', vel_array, fmt='%.6f', header='BF velocities, shape (nspec, nvel)')
            np.savetxt(location + filename_bulk +'_order_'+str(j) + '_bfvals_B.txt', bf_array, fmt='%.6f',
                       header='BF vals, shape (nspec, nvel)')
            np.savetxt(location + filename_bulk +'_order_'+str(j) + '_bfsmooth_B.txt', bf_smooth_array, fmt='%.6f',
                       header='Smoothed BF vals, shape (nspec, nvel)')
            np.savetxt(location + filename_bulk +'_order_'+str(j) + '_models_B.txt', model_array, fmt='%.6f',
                       header='BF Fit vals, shape (nspec, nvel)')
    else:
        vel_array = np.empty((bf_fitres_B.size, bf_fitres_B[0][1].size))
        bf_array = np.empty((bf_fitres_B.size, bf_fitres_B[0][1].size))
        bf_smooth_array = np.empty((bf_fitres_B.size, bf_fitres_B[0][1].size))
        model_array = np.empty((bf_fitres_B.size, bf_fitres_B[0][1].size))
        for i in range(0, bf_fitres_B.size):
            model_vals_B, bf_velocity_B, bf_vals_B, bf_smooth_vals_B = bf_fitres_B[i][1:]
            vel_array[i, :] = bf_velocity_B
            bf_array[i, :] = bf_vals_B
            bf_smooth_array[i, :] = bf_smooth_vals_B
            model_array[i, :] = model_vals_B
        np.savetxt(location + filename_bulk + '_velocities_B.txt', vel_array, fmt='%.6f', header='BF velocities, shape (nspec, nvel)')
        np.savetxt(location + filename_bulk + '_bfvals_B.txt', bf_array, fmt='%.6f', header='BF vals, shape (nspec, nvel)')
        np.savetxt(location + filename_bulk + '_bfsmooth_B.txt', bf_smooth_array, fmt='%.6f',
               header='Smoothed BF vals, shape (nspec, nvel)')
        np.savetxt(location + filename_bulk + '_models_B.txt', model_array, fmt='%.6f', header='BF Fit vals, shape (nspec, nvel)')


def _create_wavelength_intervals(
        wavelength, wavelength_intervals: float or list, flux_collection, flux_templateA, flux_templateB,
        wavelength_buffer_size, separated_flux_A: np.ndarray = None, separated_flux_B: np.ndarray = None
):
    wavelength_interval_collection = []
    flux_interval_collection = []
    templateA_interval_collection = []
    templateB_interval_collection = []
    interval_buffer_mask = []
    separated_A_interval_collection = []
    separated_B_interval_collection = []
    if wavelength_buffer_size is None:
        wavelength_buffer_size = 0.0
    w_interval_start = wavelength[0] + wavelength_buffer_size
    i = 0
    while True:
        if isinstance(wavelength_intervals, float):
            if w_interval_start + wavelength_intervals > wavelength[-1] - wavelength_buffer_size:
                w_interval_end = wavelength[-1] - wavelength_buffer_size
            else:
                w_interval_end = w_interval_start + wavelength_intervals

            if w_interval_end - w_interval_start < wavelength_intervals // 2:
                break
        else:
            if i >= len(wavelength_intervals):
                break
            w_interval_start = wavelength_intervals[i][0]
            w_interval_end = wavelength_intervals[i][1]

        w_interval = (w_interval_start, w_interval_end)

        if separated_flux_A is not None and separated_flux_B is not None:
            _, _, wl_buffered, flux_buffered_list, buffer_mask = spf.limit_wavelength_interval_multiple_spectra(
                w_interval, wavelength, flux_collection, flux_templateA, flux_templateB, separated_flux_A,
                separated_flux_B, buffer_size=wavelength_buffer_size, even_length=True
            )
            [fl_buffered, flA_buffered, flB_buffered, sflA_buffered, sflB_buffered] = flux_buffered_list
            separated_A_interval_collection.append(sflA_buffered)
            separated_B_interval_collection.append(sflB_buffered)
        else:
            _, _, wl_buffered, flux_buffered_list, buffer_mask = spf.limit_wavelength_interval_multiple_spectra(
                w_interval, wavelength, flux_collection, flux_templateA, flux_templateB,
                buffer_size=wavelength_buffer_size, even_length=True
            )
            [fl_buffered, flA_buffered, flB_buffered] = flux_buffered_list

        wavelength_interval_collection.append(wl_buffered)
        flux_interval_collection.append(fl_buffered)
        templateA_interval_collection.append(flA_buffered)
        templateB_interval_collection.append(flB_buffered)
        interval_buffer_mask.append(buffer_mask)
        i += 1
        if isinstance(wavelength_intervals, float):
            w_interval_start = w_interval_end

    if separated_flux_A is not None and separated_flux_B is not None:
        return (wavelength_interval_collection, flux_interval_collection, templateA_interval_collection,
                templateB_interval_collection, separated_A_interval_collection, separated_B_interval_collection,
                interval_buffer_mask)
    else:
        return (wavelength_interval_collection, flux_interval_collection, templateA_interval_collection,
                templateB_interval_collection, interval_buffer_mask)


def _combine_intervals(combine_intervals, wl_intervals, fl_intervals, templA_intervals, templB_intervals,
                       interval_buffer):
    for arg in reversed(combine_intervals):
        first_idx, last_idx = arg[0], arg[1]
        wl_intervals[first_idx] = np.append(wl_intervals[first_idx], wl_intervals[last_idx])
        fl_intervals[first_idx] = np.append(fl_intervals[first_idx], fl_intervals[last_idx],
                                                  axis=0)
        templA_intervals[first_idx] = np.append(templA_intervals[first_idx], templA_intervals[last_idx])
        templB_intervals[first_idx] = np.append(templB_intervals[first_idx], templB_intervals[last_idx])
        interval_buffer[first_idx] = np.append(interval_buffer[first_idx], interval_buffer[last_idx])

        del wl_intervals[last_idx]
        del fl_intervals[last_idx]
        del templA_intervals[last_idx]
        del templB_intervals[last_idx]
        del interval_buffer[last_idx]

    return wl_intervals, fl_intervals, templA_intervals, templB_intervals, interval_buffer


def spectral_separation_routine_multiple_intervals(
        wavelength: np.ndarray,
        wavelength_intervals: list,
        flux_collection: np.ndarray,
        flux_templateA: np.ndarray, flux_templateB: np.ndarray,
        RV_guess_collection: np.ndarray,
        options: RoutineOptions,
        sep_comp_options: SeparateComponentsOptions,
        rv_options: RadialVelocityOptions,
        buffer_size: float = None,
        time_values: np.ndarray = None,
        period: float = None
):
    (wl_interval_coll, flux_interval_coll, templA_interval_coll, templB_interval_coll,
     interval_buffer_mask) = _create_wavelength_intervals(
        wavelength, wavelength_intervals, flux_collection, flux_templateA, flux_templateB, buffer_size
    )

    results = []
    for i in range(0, len(wl_interval_coll)):
        wavelength_ = wl_interval_coll[i]
        flux_collection_ = flux_interval_coll[i]
        templateA_ = templA_interval_coll[i]
        templateB_ = templB_interval_coll[i]
        buffer_mask_ = interval_buffer_mask[i]

        RV_A, RV_B, sep_flux_A, sep_flux_B, wl_temp = spectral_separation_routine(
            flux_collection_, templateA_, templateB_,
            wavelength_, options, sep_comp_options, rv_options, RV_guess_collection, time_values=time_values,
            period=period, buffer_mask=buffer_mask_
        )
        results.append([RV_A, RV_B, sep_flux_A, sep_flux_B, wl_temp])
    return results

